"""Test the geoloc module."""


import datetime as dt

import numpy as np
import pytest

from pyorbital.geoloc import ScanGeometry, geodetic_lat, qrotate, subpoint
from pyorbital.geoloc_avhrr import (
    compute_avhrr_gcps_lonlatalt,
    estimate_time_and_attitude_deviations,
    estimate_time_offset,
)
from pyorbital.geoloc_instrument_definitions import (
    amsua,
    ascat,
    atms,
    avhrr,
    avhrr_from_times,
    avhrr_gac_from_times,
    hirs4,
    mhs,
    slstr_nadir,
    viirs,
)


class TestQuaternion:
    """Test the quaternion rotation."""

    def test_qrotate(self):
        """Test quaternion rotation."""
        vector = np.array([[1, 0, 0]]).T
        axis = np.array([[0, 1, 0]]).T
        angle = np.deg2rad(90)

        result = qrotate(vector, axis, angle)[:, 0]
        expected = np.array([0, 0, 1])
        np.testing.assert_allclose(result, expected, rtol=1e-8, atol=1e-8)

        axis = np.array([0, 1, 0])
        result = qrotate(vector, axis, angle)
        expected = np.array([[0, 0, 1]]).T
        np.testing.assert_allclose(result, expected, rtol=1e-8, atol=1e-8)

        vector = np.array([[1, 0, 0],
                           [0, 0, 1]]).T
        axis = np.array([0, 1, 0])
        angle = np.deg2rad(90)
        result = qrotate(vector, axis, angle)
        expected = np.array([[0, 0, 1],
                             [-1, 0, 0]]).T

        np.testing.assert_allclose(result, expected, rtol=1e-8, atol=1e-8)

        axis = np.array([[0, 1, 0]]).T
        result = qrotate(vector, axis, angle)
        expected = np.array([[0, 0, 1],
                             [-1, 0, 0]]).T

        np.testing.assert_allclose(result, expected, rtol=1e-8, atol=1e-8)


class TestGeoloc:
    """Test for the core computing part."""

    def test_scan_geometry(self):
        """Test the ScanGeometry object."""
        scans_nb = 1

        xy = np.vstack((np.deg2rad(np.array([10, 0, -10])),
                        np.array([0, 0, 0])))
        xy = np.tile(xy[:, np.newaxis, :], [1, np.int32(scans_nb), 1])

        times = np.tile([-0.1, 0, 0.1], [np.int32(scans_nb), 1])

        instrument = ScanGeometry(xy, times)

        np.testing.assert_allclose(np.rad2deg(instrument.fovs[0]), np.array([[10, 0, -10]]))

        # Test vectors

        pos = np.rollaxis(np.tile(np.array([0, 0, 7000]), [3, 1, 1]), 2)
        vel = np.rollaxis(np.tile(np.array([1, 0, 0]), [3, 1, 1]), 2)
        pos = np.stack([np.array([0, 0, 7000])] * 3, 1)[:, np.newaxis, :]
        vel = np.stack([np.array([1, 0, 0])] * 3, 1)[:, np.newaxis, :]

        vec = instrument.vectors(pos, vel)

        result = vec[:, 0, 1]
        expected = np.array([0.0, 0.0, -1.0])
        np.testing.assert_allclose(result, expected, rtol=1e-8, atol=1e-8)

        # Check if we can pass an array for yaw
        vec = instrument.vectors(pos, vel, yaw=[0])

        result = vec[:, 0, 1]
        expected = np.array([0.0, 0.0, -1.0])
        np.testing.assert_allclose(result, expected, rtol=1e-8, atol=1e-8)


        # minus sin because we use trigonometrical direction of angles
        result = vec[:, 0, 0]
        expected = np.array([0, -np.sin(np.deg2rad(10)), -np.cos(np.deg2rad(10))])
        np.testing.assert_allclose(result, expected, rtol=1e-7, atol=1e-7)

        result = vec[:, 0, 2]
        expected = np.array([0, -np.sin(np.deg2rad(-10)), -np.cos(np.deg2rad(-10))])
        np.testing.assert_allclose(result, expected, rtol=1e-7, atol=1e-7)

        # Test times

        start_of_scan = np.datetime64(dt.datetime(2014, 1, 8, 11, 30))
        times = instrument.times(start_of_scan)

        assert times[0, 1] == start_of_scan
        assert times[0, 0] == start_of_scan - np.timedelta64(100, "ms")
        assert times[0, 2] == start_of_scan + np.timedelta64(100, "ms")

    def test_geodetic_lat(self):
        """Test the determination of the geodetic latitude."""
        point = np.array([[7000, 0, 7000]]).T
        np.testing.assert_allclose(geodetic_lat(point),
                                   np.array([0.78755832699854733]), rtol=1e-8, atol=1e-8)

        points = np.array([[7000, 0, 7000],
                           [7000, 0, 7000]]).T
        result = geodetic_lat(points)
        expected = np.array([0.78755832699854733, 0.78755832699854733])
        np.testing.assert_allclose(result, expected, rtol=1e-8, atol=1e-8)

    def test_subpoint(self):
        """Test nadir determination."""
        a = 6378.137  # km
        b = 6356.75231414  # km, GRS80
        point = np.array([0, 0, 7000])
        nadir = subpoint(point, a, b)
        np.testing.assert_allclose(nadir, np.array([0, 0, b]), rtol=1e-7, atol=1e-7)

        point = np.array([7000, 0, 7000])
        nadir = subpoint(point, a, b)
        np.testing.assert_allclose(nadir,
                                   np.array([4507.85431429,
                                             0,
                                             4497.06396339]), rtol=1e-8, atol=1e-8)
        points = np.array([[7000, 0, 7000],
                           [7000, 0, 7000]]).T
        nadir = subpoint(points, a, b)
        np.testing.assert_allclose(nadir[:, 0],
                                   np.array([4507.85431429,
                                             0,
                                             4497.06396339]), rtol=1e-8, atol=1e-8)
        np.testing.assert_allclose(nadir[:, 1],
                                   np.array([4507.85431429,
                                             0,
                                             4497.06396339]), rtol=1e-8, atol=1e-8)




def test_arbitrary_point_geoloc():
    """Test geolocating an arbitrary point in the swath."""
    from pyorbital.geoloc_avhrr import compute_avhrr_gcps_lonlatalt

    # Couple of example Two Line Elements
    tle1 = "1 33591U 09005A   12345.45213434  .00000391  00000-0  24004-3 0  6113"
    tle2 = "2 33591 098.8821 283.2036 0013384 242.4835 117.4960 14.11432063197875"

    # Choosing a specific time, this should be relatively close to the issue date of the TLE
    t = dt.datetime(2012, 12, 12, 4, 16, 1, 575000)
    rpy = (0, 0, 0)

    max_scan_angle = 55.37

    gcps = np.array([[2, 500], [1500, 700], [20, 1000]])

    lons, lats, alts = compute_avhrr_gcps_lonlatalt(gcps, max_scan_angle, rpy, t, (tle1, tle2))

    assert lons[0] == pytest.approx(-34.69996894)
    assert lats[0] == pytest.approx(56.69799502)

    assert lons[2] == pytest.approx(-27.573052737698944)
    assert lats[2] == pytest.approx(55.626740897592654)


def test_minimize_geoloc_error():
    """Test minimizing the distance to a set of gcps."""
    # Couple of example Two Line Elements
    tle1 = "1 33591U 09005A   12345.45213434  .00000391  00000-0  24004-3 0  6113"
    tle2 = "2 33591 098.8821 283.2036 0013384 242.4835 117.4960 14.11432063197875"
    tle = (tle1, tle2)

    # Choosing a specific time, this should be relatively close to the issue date of the TLE
    t = dt.datetime(2012, 12, 12, 4, 16, 1, 575000)

    ref_time_displacement = 0.51
    ref_time = t + dt.timedelta(seconds=ref_time_displacement)
    ref_yaw = 0.1
    rpy = (0, 0, ref_yaw)
    max_scan_angle = 55.37
    # gcps are line/col
    gcps = np.array([[2, 500], [1500, 700], [20, 1000], [500, 1100], [100, 2000]])
    ref_lons, ref_lats, _ = compute_avhrr_gcps_lonlatalt(gcps, max_scan_angle, rpy, ref_time, tle)
    time_diff, (roll, pitch, yaw), (do, dm) = estimate_time_and_attitude_deviations(gcps, ref_lons, ref_lats, t,
                                                                                    tle, max_scan_angle)
    assert time_diff == pytest.approx(ref_time_displacement, abs=1e-2)
    assert yaw == pytest.approx(ref_yaw, abs=1e-2)
    assert min(do) > max(dm)


def test_minimize_time_error():
    """Test minimizing the distance to a set of gcps using only time offset."""
    # Couple of example Two Line Elements
    tle1 = "1 33591U 09005A   12345.45213434  .00000391  00000-0  24004-3 0  6113"
    tle2 = "2 33591 098.8821 283.2036 0013384 242.4835 117.4960 14.11432063197875"
    tle = (tle1, tle2)

    # Choosing a specific time, this should be relatively close to the issue date of the TLE
    t = dt.datetime(2012, 12, 12, 4, 16, 1, 575000)

    ref_time_displacement = 20
    ref_time = t + dt.timedelta(seconds=ref_time_displacement)
    rpy = (0, 0, 0)
    max_scan_angle = 55.37
    # gcps are line/col
    gcps = np.array([[2, 500], [1500, 700], [20, 1000], [500, 1100], [100, 2000]])
    ref_lons, ref_lats, _ = compute_avhrr_gcps_lonlatalt(gcps, max_scan_angle, rpy, ref_time, tle)
    time_diff, (do, dm) = estimate_time_offset(gcps, ref_lons, ref_lats, t,
                                                                      tle, max_scan_angle)
    assert time_diff == pytest.approx(ref_time_displacement, abs=1e-2)
    assert min(do) > max(dm)


class TestGeolocDefs:
    """Test the instrument definitions."""

    def test_avhrr(self):
        """Test the definition of the avhrr instrument."""
        avh = avhrr(1, np.array([0, 1023.5, 2047]))
        result = np.rad2deg(avh.fovs[0])
        expected = np.array([[55.37, 0, -55.37]])
        np.testing.assert_allclose(result, expected, rtol=1e-7, atol=1e-7)

        avh = avhrr(1, np.array([0, 1023.5, 2047]), 10)
        np.testing.assert_allclose(np.rad2deg(avh.fovs[0]),
                                   np.array([[10, 0, -10]]))

        # This is perhaps a bit odd, to require avhrr to accept floats for
        # the number of scans? FIXME!
        avh = avhrr(1.1, np.array([0, 1023.5, 2047]), 10)
        np.testing.assert_allclose(np.rad2deg(avh.fovs[0]),
                                   np.array([[10, 0, -10]]))

    def test_avhrr_from_times(self):
        """Test generating the avhrr from times."""
        avh = avhrr_from_times([dt.datetime(2000,1,1,0,0,0)], [0, 1023.5, 2047])
        result = np.rad2deg(avh.fovs[0])
        expected = np.array([[55.37, 0, -55.37]])
        np.testing.assert_allclose(result, expected, rtol=1e-7, atol=1e-7)
        result = avh.times(dt.date(2000,1,1))
        expected = ((np.array([[0, 1023.5, 2047]]) * 25000).astype("timedelta64[ns]")
                    + np.datetime64("2000-01-01T00:00:00"))
        np.testing.assert_equal(result, expected)

        avh = avhrr_from_times([dt.datetime(2000,1,1,0,0,0)], np.array([0, 1023.5, 2047]), 10)
        np.testing.assert_allclose(np.rad2deg(avh.fovs[0]),
                                   np.array([[10, 0, -10]]))

        avh = avhrr_from_times([dt.datetime(2000,1,1,0,0,0), dt.datetime(2000,1,1,0,1,0)],
                    [0, 2047])
        times = avh.times(dt.datetime(2001,1,1))
        expected = (np.array([[0,51175000],[60000000000, 60051175000]]).astype("timedelta64[ns]")
                    + np.datetime64("2001-01-01"))
        np.testing.assert_equal(times, expected)


    def test_avhrr_gac_from_times(self):
        """Test getting avhrr gac from times."""
        avh = avhrr_gac_from_times([dt.datetime(2000,1,1,0,0,0)], [0, 204, 408])
        result = np.rad2deg(avh.fovs[0])
        expected = np.array([[55.180655, 0, -55.180655]])
        np.testing.assert_allclose(result, expected, rtol=1e-7, atol=1e-7)
        result = avh.times(dt.date(2000,1,1))
        expected = ((np.array([[0, 204, 408]]) * 125000).astype("timedelta64[ns]")
                    + np.datetime64("2000-01-01T00:00:00"))
        np.testing.assert_equal(result, expected)

        avh = avhrr_gac_from_times([dt.datetime(2000,1,1,0,0,0)], np.array([0, 204, 408]), 10)
        np.testing.assert_allclose(np.rad2deg(avh.fovs[0]),
                                   np.array([[9.965804, 0, -9.965804]]))

        avh = avhrr_gac_from_times([dt.datetime(2000,1,1,0,0,0), dt.datetime(2000,1,1,0,1,0)],
                    [0, 408])
        times = avh.times(dt.datetime(2001,1,1))
        expected = (np.array([[0,51000000],[60000000000, 60051000000]]).astype("timedelta64[ns]")
                    + np.datetime64("2001-01-01"))
        np.testing.assert_equal(times, expected)


    def test_viirs(self):
        """Test the definition of the viirs instrument."""
        geom = viirs(1, np.array([0, 3200, 6399]))
        expected_fovs = np.array([
            np.tile(np.array([[0.98, -0., -0.98]]), [32, 1]),
            np.tile(np.array([[0., -0., 0]]), [32, 1])], dtype=np.float64)

        np.testing.assert_allclose(geom.fovs,
                                   expected_fovs, rtol=1e-2, atol=1e-2)

        geom = viirs(2, np.array([0, 3200, 6399]))
        expected_fovs = np.array([
            np.tile(np.array([[0.98, -0., -0.98]]), [32*2, 1]),
            np.tile(np.array([[0., -0., 0]]), [32*2, 1])], dtype=np.float64)

        np.testing.assert_allclose(geom.fovs,
                                   expected_fovs, rtol=1e-2, atol=1e-2)

    def test_viirs_defaults(self):
        """Test the definition of the viirs instrument with default slicing."""
        geom = viirs(1, chn_pixels=3)
        expected_fovs = np.array([
            np.tile(np.array([[0.98, -0., -0.98]]), [32, 1]),
            np.tile(np.array([[0., -0., 0]]), [32, 1])], dtype=np.float64)

        np.testing.assert_allclose(geom.fovs,
                                   expected_fovs, rtol=1e-2, atol=1e-2)

    def test_amsua(self):
        """Test the definition of the amsua instrument."""
        geom = amsua(1)
        expected_fovs = np.array([
            [[0.84,  0.78,  0.73,  0.67,  0.61,  0.55,  0.49,  0.44,  0.38,
              0.32,  0.26,  0.2,  0.15,  0.09,  0.03, -0.03, -0.09, -0.15,
              -0.2, -0.26, -0.32, -0.38, -0.44, -0.49, -0.55, -0.61, -0.67,
              -0.73, -0.78, -0.84]],
            np.zeros((1, 30))], dtype=np.float64)
        np.testing.assert_allclose(geom.fovs, expected_fovs, rtol=1e-2, atol=1e-2)

    def test_mhs(self):
        """Test the definition of the mhs instrument."""
        geom = mhs(1)
        expected_fovs = np.array([
            [[0.86,  0.84,  0.82,  0.8,  0.79,  0.77,  0.75,  0.73,  0.71,
              0.69,  0.67,  0.65,  0.63,  0.61,  0.59,  0.57,  0.55,  0.53,
              0.51,  0.49,  0.48,  0.46,  0.44,  0.42,  0.4,  0.38,  0.36,
              0.34,  0.32,  0.3,  0.28,  0.26,  0.24,  0.22,  0.2,  0.18,
              0.16,  0.15,  0.13,  0.11,  0.09,  0.07,  0.05,  0.03,  0.01,
              -0.01, -0.03, -0.05, -0.07, -0.09, -0.11, -0.13, -0.15, -0.16,
              -0.18, -0.2, -0.22, -0.24, -0.26, -0.28, -0.3, -0.32, -0.34,
              -0.36, -0.38, -0.4, -0.42, -0.44, -0.46, -0.48, -0.49, -0.51,
              -0.53, -0.55, -0.57, -0.59, -0.61, -0.63, -0.65, -0.67, -0.69,
              -0.71, -0.73, -0.75, -0.77, -0.79, -0.8, -0.82, -0.84, -0.86]],
            np.zeros((1, 90))], dtype=np.float64)
        np.testing.assert_allclose(geom.fovs,
                                   expected_fovs, rtol=1e-2, atol=1e-2)

    def test_hirs4(self):
        """Test the definition of the hirs4 instrument."""
        geom = hirs4(1)
        expected_fovs = np.array([
            [[0.86,  0.83,  0.8,  0.77,  0.74,  0.71,  0.68,  0.64,  0.61,
              0.58,  0.55,  0.52,  0.49,  0.46,  0.42,  0.39,  0.36,  0.33,
              0.3,  0.27,  0.24,  0.2,  0.17,  0.14,  0.11,  0.08,  0.05,
              0.02, -0.02, -0.05, -0.08, -0.11, -0.14, -0.17, -0.2, -0.24,
              -0.27, -0.3, -0.33, -0.36, -0.39, -0.42, -0.46, -0.49, -0.52,
              -0.55, -0.58, -0.61, -0.64, -0.68, -0.71, -0.74, -0.77, -0.8,
              -0.83, -0.86]],
            np.zeros((1, 56))], dtype=np.float64)
        np.testing.assert_allclose(geom.fovs,
                                   expected_fovs, rtol=1e-2, atol=1e-2)

    def test_atms(self):
        """Test the definition of the atms instrument."""
        geom = atms(1)
        expected_fovs = np.array([
            [[0.92,  0.9,  0.88,  0.86,  0.84,  0.82,  0.8,  0.78,  0.76,
              0.75,  0.73,  0.71,  0.69,  0.67,  0.65,  0.63,  0.61,  0.59,
              0.57,  0.55,  0.53,  0.51,  0.49,  0.47,  0.46,  0.44,  0.42,
              0.4,  0.38,  0.36,  0.34,  0.32,  0.3,  0.28,  0.26,  0.24,
              0.22,  0.2,  0.18,  0.16,  0.15,  0.13,  0.11,  0.09,  0.07,
              0.05,  0.03,  0.01, -0.01, -0.03, -0.05, -0.07, -0.09, -0.11,
              -0.13, -0.15, -0.16, -0.18, -0.2, -0.22, -0.24, -0.26, -0.28,
              -0.3, -0.32, -0.34, -0.36, -0.38, -0.4, -0.42, -0.44, -0.46,
              -0.47, -0.49, -0.51, -0.53, -0.55, -0.57, -0.59, -0.61, -0.63,
              -0.65, -0.67, -0.69, -0.71, -0.73, -0.75, -0.76, -0.78, -0.8,
              -0.82, -0.84, -0.86, -0.88, -0.9, -0.92]],
            np.zeros((1, 96))], dtype=np.float64)
        np.testing.assert_allclose(geom.fovs,
                                   expected_fovs, rtol=1e-2, atol=1e-2)

    def test_ascat(self):
        """Test the definition of the ASCAT instrument onboard Metop."""
        geom = ascat(1)
        expected_fovs = np.array([
            [[0.9250245,  0.90058989,  0.87615528,  0.85172067,
              0.82728607,  0.80285146,  0.77841685,  0.75398224,
              0.72954763,  0.70511302,  0.68067841,  0.6562438,
              0.63180919,  0.60737458,  0.58293997,  0.55850536,
              0.53407075,  0.50963614,  0.48520153,  0.46076692,
              0.43633231, -0.43633231, -0.46076692, -0.48520153,
              -0.50963614, -0.53407075, -0.55850536, -0.58293997,
              -0.60737458, -0.63180919, -0.6562438, -0.68067841,
              -0.70511302, -0.72954763, -0.75398224, -0.77841685,
              -0.80285146, -0.82728607, -0.85172067, -0.87615528,
              -0.90058989, -0.9250245]], np.zeros((1, 42))], dtype=np.float64)

        np.testing.assert_allclose(
            geom.fovs, expected_fovs, rtol=1e-2, atol=1e-2)
        geom = ascat(1, np.array([0, 41]))
        expected_fovs = np.array([[[0.9250245,  -0.9250245]],
                                  [[0.,  0.]]], dtype=np.float64)
        np.testing.assert_allclose(
            geom.fovs, expected_fovs, rtol=1e-2, atol=1e-2)

        geom = ascat(1, np.array([0, -1]))
        np.testing.assert_allclose(
            geom.fovs, expected_fovs, rtol=1e-2, atol=1e-2)

    def test_slstr_nadir(self):
        """Test the definition of the slstr instrument nadir view flying on Sentinel-3."""
        geom = slstr_nadir(1, [0, 1])

        expected_fovs = np.array([
            np.tile(np.array([[0.8115781, -0.38571776]]), [1, 1]),
            np.tile(np.array([[0., 0.]]), [1, 1])], dtype=np.float64)
        np.testing.assert_allclose(geom.fovs, expected_fovs, rtol=1e-2, atol=1e-2)

        geom = slstr_nadir(1, None)

        np.testing.assert_equal(geom.fovs.size, 6000)



"""This module provides geoloc operations specific to the avhrr instrument.

In particular, it provides functions that allow matching gcp location in swath coordinates to reference positions, and
then minimise the distance to these positions by adjusting the time offset and attitude.
"""

import logging

import numpy as np
from pyproj import Geod

from pyorbital.geoloc import ScanGeometry, compute_pixels, get_lonlatalt

logger = logging.getLogger(__name__)
geod = Geod(ellps="WGS84")

def compute_avhrr_gcps_lonlatalt(gcps, max_scan_angle, rpy, start_time, tle) -> None:
    """Compute the longitute, latitude and altitude of given gcps (scanlines, columns of the swath).

    The gcps are arbitrary location in swath coordinates, for example (10.3, 7.7) for a gcp at line 10.3 in the swath,
    and column 7.7. This function returns the geographical coordinates of the gcps.

    The scanlines are relative to the pass scanline numbers, zero-based.
    """
    time_line_interval = 1/6
    time_row_interval = 25e-6

    fov_x = gcps[:, 1]
    fov_y = gcps[:, 0]

    scan_angles_across = (fov_x / 1023.5 - 1) * np.deg2rad(-max_scan_angle)
    scan_angles_along = np.zeros_like(scan_angles_across)
    scan_angles = np.vstack((scan_angles_across, scan_angles_along))
    time_offsets = np.array(fov_x * time_row_interval + fov_y * time_line_interval)
    geom = ScanGeometry(scan_angles, time_offsets)
    start_time = np.datetime64(start_time)
    s_times = geom.times(start_time)

    pixels_pos = compute_pixels(tle, geom, s_times, rpy)
    return get_lonlatalt(pixels_pos, s_times)


def estimate_time_and_attitude_deviations(gcps, ref_lons, ref_lats, start_time, tle, max_scan_angle):
    """Estimate time offset and attitude deviations from gcps.

    Provided reference longitudes and latitudes for the gcps, this function minimises the attitude and time offset
    needed to match the gcp coordinates to the reference coordinates.
    """
    from scipy.optimize import minimize

    original_distances = compute_gcp_distances_to_reference_lonlats((0, 0, 0, 0), gcps, start_time, tle, max_scan_angle,
                                                                    (ref_lons, ref_lats))
    original_median_distance = np.median(original_distances)
    logger.debug(f"GCP distances: median {original_median_distance}, std {np.std(original_distances)}")
    # we need to work in seconds*1e3 to avoid the nanosecond precision issue
    res = minimize(compute_gcp_accumulated_squared_distances_to_reference_lonlats,
                   x0=(0, 0, 0, 0),
                   args=(gcps, start_time, tle, max_scan_angle, (ref_lons, ref_lats)),
                   bounds=((-0.007, 0.007) , (-0.5, 0.5), (-0.5, 0.5), (-0.5, 0.5)))
    if not res.success:
        raise RuntimeError("Time and attitude estimation did not converge")
    time_diff, roll, pitch, yaw = res.x * [1e3, 1, 1, 1]
    logger.debug(f"Estimated time difference to {time_diff} seconds, "
                 f"attitude to {np.rad2deg(roll)}, {np.rad2deg(pitch)}, {np.rad2deg(yaw)} degrees")
    distances = compute_gcp_distances_to_reference_lonlats(res.x, gcps, start_time, tle, max_scan_angle,
                                                           (ref_lons, ref_lats))

    minimized_median_distance = np.median(distances)
    logger.debug(f"Remaining GCP distances: median {minimized_median_distance}, std {np.std(distances)}")

    return time_diff, (roll, pitch, yaw), (original_distances, distances)


def estimate_time_offset(gcps, ref_lons, ref_lats, start_time, tle, max_scan_angle):
    """Estimate time offset from gcps.

    Provided reference longitudes and latitudes for the gcps, this function minimises the time offset
    needed to match the gcp coordinates to the reference coordinates.
    """
    from scipy.optimize import minimize

    original_distances = compute_gcp_distances_to_reference_lonlats((0, 0, 0, 0), gcps, start_time, tle, max_scan_angle,
                                                                    (ref_lons, ref_lats))
    original_median_distance = np.median(original_distances)
    logger.debug(f"GCP distances: median {original_median_distance}, std {np.std(original_distances)}")

    def gcp_distance_for_time(time):
        dist = compute_gcp_accumulated_squared_distances_to_reference_lonlats((time[0], 0, 0, 0), gcps, start_time, tle,
                                                                              max_scan_angle, (ref_lons, ref_lats))
        return dist

    # we need to work in seconds*1e3 to avoid the nanosecond precision issue
    res = minimize(gcp_distance_for_time,
                   x0=(0,),
                   bounds=((-0.03, 0.03),),
                   options=dict(ftol=1e-1),
                   )
    if not res.success:
        raise RuntimeError("Time offset estimation did not converge")
    time_diff, = res.x * [1e3,]
    logger.debug(f"Estimated time difference to {time_diff} seconds")
    distances = compute_gcp_distances_to_reference_lonlats((res.x[0], 0, 0, 0), gcps, start_time, tle, max_scan_angle,
                                                           (ref_lons, ref_lats))

    minimized_median_distance = np.median(distances)
    logger.debug(f"Remaining GCP distances: median {minimized_median_distance}, std {np.std(distances)}")

    return time_diff, (original_distances, distances)


def compute_gcp_accumulated_squared_distances_to_reference_lonlats(
        variables, gcps, start_time, tle, max_scan_angle, refs):
    """Compute the summed squared distance fot gcps to reference lonlats.

    Given the gcps (in swath coordinates) along with attitude and time offset, compute the sum of squared distances to
    the reference lons and lats of the gcps.
    """
    distances = compute_gcp_distances_to_reference_lonlats(variables, gcps, start_time, tle, max_scan_angle, refs)
    return np.sum(distances**2)


def compute_gcp_distances_to_reference_lonlats(variables, gcps, start_time, tle, max_scan_angle, refs):
    """Compute the gcp distances to references lonlats."""
    time_diff, roll, pitch, yaw = variables
    # we need to work in seconds*1e3 to avoid the nanosecond precision issue
    time = np.datetime64(start_time) + np.timedelta64(int(time_diff * 1e12), "ns")
    lons, lats, _ = compute_avhrr_gcps_lonlatalt(gcps, max_scan_angle, (roll, pitch, yaw), time, tle)
    valid = np.isfinite(lons)
    lons = lons[valid]
    lats = lats[valid]
    ref_lons, ref_lats = refs
    ref_lons = np.array(ref_lons)[valid]
    ref_lats = np.array(ref_lats)[valid]
    _, _, distances = geod.inv(ref_lons, ref_lats, lons, lats)
    return distances



"""Angle and time-based astronomy functions.

Parts taken from http://www.geoastro.de/elevaz/basics/index.htm

Note on argument types
----------------------

Many of these functions accept Python datetime objects,
numpy datetime64 objects, or anything that can be turned
into a numpy array of datetime64 objects. These objects are inherently
64-bit so if other arguments (ex. longitude and latitude arrays) are
32-bit floats internal operations will be automatically promoted to
64-bit floating point numbers. Where possible these are then converted
back to 32-bit before being returned. In general scalar inputs will also
produce scalar outputs.

"""

import numpy as np

from pyorbital import dt2np

F = 1 / 298.257223563  # Earth flattening WGS-84
A = 6378.137  # WGS84 Equatorial radius
MFACTOR = 7.292115E-5


def jdays2000(utc_time):
    """Get the days since year 2000."""
    return _days(dt2np(utc_time) - np.datetime64("2000-01-01T12:00"))


def jdays(utc_time):
    """Get the julian day of *utc_time*."""
    return jdays2000(utc_time) + 2451545.0


def _days(dt):
    """Get the days (floating point) from *d_t*."""
    if hasattr(dt, "shape"):
        dt = np.asanyarray(dt, dtype=np.timedelta64)
    return dt / np.timedelta64(1, "D")


def gmst(utc_time):
    """Greenwich mean sidereal utc_time, in radians.

    As defined in the AIAA 2006 implementation:
    http://www.celestrak.com/publications/AIAA/2006-6753/
    """
    ut1 = jdays2000(utc_time) / 36525.0
    theta = 67310.54841 + ut1 * (876600 * 3600 + 8640184.812866 + ut1 *
                                 (0.093104 - ut1 * 6.2 * 10e-6))
    return np.deg2rad(theta / 240.0) % (2 * np.pi)


def _lmst(utc_time, longitude):
    """Local mean sidereal time, computed from *utc_time* and *longitude*.

    utc_time: The UTC time as a datetime.datetime  object.
    Logitude: The longitude in radians.
    Returns: local mean sideral time in radians.
    """
    return gmst(utc_time) + longitude


def sun_ecliptic_longitude(utc_time):
    """Ecliptic longitude of the sun at *utc_time*."""
    jdate = jdays2000(utc_time) / 36525.0
    # mean anomaly, rad
    m_a = np.deg2rad(357.52910 +
                     35999.05030 * jdate -
                     0.0001559 * jdate * jdate -
                     0.00000048 * jdate * jdate * jdate)
    # mean longitude, deg
    l_0 = 280.46645 + 36000.76983 * jdate + 0.0003032 * jdate * jdate
    d_l = ((1.914600 - 0.004817 * jdate - 0.000014 * jdate * jdate) * np.sin(m_a) +
           (0.019993 - 0.000101 * jdate) * np.sin(2 * m_a) + 0.000290 * np.sin(3 * m_a))
    # true longitude, deg
    l__ = l_0 + d_l
    return np.deg2rad(l__)


def sun_ra_dec(utc_time):
    """Right ascension and declination of the sun at *utc_time*."""
    jdate = jdays2000(utc_time) / 36525.0
    eps = np.deg2rad(23.0 + 26.0 / 60.0 + 21.448 / 3600.0 -
                     (46.8150 * jdate + 0.00059 * jdate * jdate -
                      0.001813 * jdate * jdate * jdate) / 3600)
    eclon = sun_ecliptic_longitude(utc_time)
    x__ = np.cos(eclon)
    y__ = np.cos(eps) * np.sin(eclon)
    z__ = np.sin(eps) * np.sin(eclon)
    r__ = np.sqrt(1.0 - z__ * z__)
    # sun declination
    declination = np.arctan2(z__, r__)
    # right ascension
    right_ascension = 2 * np.arctan2(y__, (x__ + r__))
    return right_ascension, declination


def _local_hour_angle(utc_time, longitude, right_ascension):
    """Derive the hour angle at *utc_time* for the given *longitude* and *right_ascension*.

    utc_time: datetime.datetime instance of the UTC time
    longitude: Longitude in radians.
    right_ascension: The right ascension in radians.
    Returns: Hour angle in radians.
    """
    return _lmst(utc_time, longitude) - right_ascension


def get_alt_az(utc_time, lon, lat):
    """Return sun altitude and azimuth from *utc_time*, *lon*, and *lat*.

    lon,lat in degrees
    The returned angles are given in radians.
    """
    lon = np.deg2rad(lon)
    lat = np.deg2rad(lat)

    ra_, dec = sun_ra_dec(utc_time)
    h__ = _local_hour_angle(utc_time, lon, ra_)
    alt_az = (np.arcsin(np.sin(lat) * np.sin(dec) +
                        np.cos(lat) * np.cos(dec) * np.cos(h__)),
              np.arctan2(-np.sin(h__), (np.cos(lat) * np.tan(dec) -
                                        np.sin(lat) * np.cos(h__))))
    if not isinstance(lon, float):
        alt_az = (alt_az[0].astype(lon.dtype), alt_az[1].astype(lon.dtype))
    return alt_az


def cos_zen(utc_time, lon, lat):
    """Derive the cosine of the sun-zenith angle for *lon*, *lat* at *utc_time*.

    utc_time: datetime.datetime instance of the UTC time
    lon: Longitude in degrees
    lat: Latitude in degrees.
    Returns: Cosine of the sun zenith angle.
    """
    lon = np.deg2rad(lon)
    lat = np.deg2rad(lat)

    r_a, dec = sun_ra_dec(utc_time)
    h__ = _local_hour_angle(utc_time, lon, r_a)
    csza = (np.sin(lat) * np.sin(dec) + np.cos(lat) * np.cos(dec) * np.cos(h__))
    if not isinstance(lon, float):
        csza = csza.astype(lon.dtype)
    return csza


def sun_zenith_angle(utc_time, lon, lat):
    """Sun-zenith angle for *lon*, *lat* at *utc_time*.

    lon,lat in degrees.
    The sun zenith angle returned is in degrees.
    """
    sza = np.rad2deg(np.arccos(cos_zen(utc_time, lon, lat)))
    if not isinstance(lon, float):
        sza = sza.astype(lon.dtype)
    return sza


def sun_earth_distance_correction(utc_time):
    """Calculate the sun earth distance correction, relative to 1 AU."""
    # Computation according to
    #  https://web.archive.org/web/20150117190838/http://curious.astro.cornell.edu/question.php?number=582
    # with
    #  Astronomical unit: AU = 149597870700.0 meter (https://ssd.jpl.nasa.gov/glossary/au.html)
    #  Semi-major axis:    a = 1.00000261 AU = 149598261150.0 m (https://ssd.jpl.nasa.gov/planets/approx_pos.html)
    #  Eccentricity:       e = 0.01671123 (https://ssd.jpl.nasa.gov/planets/approx_pos.html)
    #  Length of year:  year = 365.25636 days (https://ssd.jpl.nasa.gov/astro_par.html)
    #  Perihelion:         p = 3 (day of year with Earth in perihelion, varies between Jan 2 and Jan 5,
    #                             http://www.astropixels.com/ephemeris/perap2001.html
    #                             https://web.archive.org/web/20080328044924/http://aa.usno.navy.mil/data/docs/EarthSeasons)
    # Formula:
    #  theta = (jdays2000(utc_time) - p) * (2 * np.pi) / year
    #  r = a * (1 - e * e) / (1 + e * np.cos(theta))
    #  corr := r/AU
    #        =  a * (1 - e * e) / AU / (1 + e * np.cos(theta))
    #       "=" a * (1 - e * e) / AU * (1 - e * np.cos(theta))
    #       "=" 1 - 0.0167 * np.cos(theta)

    corr = 1 - 0.0167 * np.cos(2 * np.pi * (jdays2000(utc_time) - 3) / 365.25636)
    return corr


def observer_position(utc_time, lon, lat, alt):
    """Calculate observer ECI position.

    http://celestrak.com/columns/v02n03/
    """
    lon = np.deg2rad(lon)
    lat = np.deg2rad(lat)

    theta = (gmst(utc_time) + lon) % (2 * np.pi)
    c = 1 / np.sqrt(1 + F * (F - 2) * np.sin(lat)**2)
    sq = c * (1 - F)**2

    achcp = (A * c + alt) * np.cos(lat)
    x = achcp * np.cos(theta)  # kilometers
    y = achcp * np.sin(theta)
    z = (A * sq + alt) * np.sin(lat)

    vx = -MFACTOR * y  # kilometers/second
    vy = MFACTOR * x
    vz = _float_to_sibling_result(0.0, vx)

    if not isinstance(lon, float):
        x = x.astype(lon.dtype, copy=False)
        y = y.astype(lon.dtype, copy=False)
        z = z.astype(lon.dtype, copy=False)
        vx = vx.astype(lon.dtype, copy=False)
        vy = vy.astype(lon.dtype, copy=False)
        vz = vz.astype(lon.dtype, copy=False)  # type: ignore[union-attr]
    return (x, y, z), (vx, vy, vz)


def _float_to_sibling_result(result_to_convert, template_result):
    """Convert a scalar to the same type as another return type.

    This is mostly used to make a static value consistent with the types of
    other returned values.

    """
    if isinstance(template_result, float):
        return result_to_convert
    # get any array like object that might be wrapped by our template (ex. xarray DataArray)
    array_like = template_result if hasattr(template_result, "__array_function__") else template_result.data
    array_convert = np.asarray(result_to_convert, like=array_like)
    if not hasattr(template_result, "__array_function__"):
        # the template result has some wrapper class (likely xarray DataArray)
        # recreate the wrapper object
        array_convert = template_result.__class__(array_convert)
    return array_convert



"""Classes and functions for handling TLE files."""
import contextlib
import datetime as dt
import glob
import io
import logging
import os
import sqlite3
import warnings
from itertools import zip_longest
from urllib.request import urlopen
from warnings import warn

import defusedxml.ElementTree as ET
import numpy as np
import requests

from pyorbital.config import config

TLE_GROUPS = ("active",
              "weather",
              "resource",
              "cubesat",
              "stations",
              "sarsat",
              "noaa",
              "amateur",
              "engineering")

TLE_URLS = [f"https://celestrak.org/NORAD/elements/gp.php?GROUP={group}&FORMAT=tle"
            for group in TLE_GROUPS]


LOGGER = logging.getLogger(__name__)
PKG_CONFIG_DIR = os.path.join(os.path.realpath(os.path.dirname(__file__)), "etc")


def _get_config_path():
    """Get the config path for Pyorbital."""
    if "PPP_CONFIG_DIR" in os.environ and "config_path" not in config:
        LOGGER.warning(
            "The use of PPP_CONFIG_DIR is no longer supported!" +
            " Please use PYORBITAL_CONFIG_PATH if you need a custom config path for pyorbital!")
        LOGGER.debug("Using the package default for configuration: %s", PKG_CONFIG_DIR)
        return PKG_CONFIG_DIR
    else:
        pyorbital_config_path = config.get("config_path", PKG_CONFIG_DIR)

    LOGGER.debug("Path to the Pyorbital configuration (where e.g. platforms.txt is found): %s",
                 str(pyorbital_config_path))
    return pyorbital_config_path


def get_platforms_filepath():
    """Get the platforms.txt file path.

    Check that the file exists or raise an error.
    """
    config_path = _get_config_path()
    platform_file = os.path.join(config_path, "platforms.txt")
    if not os.path.isfile(platform_file):
        platform_file = os.path.join(PKG_CONFIG_DIR, "platforms.txt")
        if not os.path.isfile(platform_file):
            raise OSError("Platform file {filepath} does not exist!".format(filepath=platform_file))

    return platform_file


def read_platform_numbers(filename, in_upper=False, num_as_int=False):
    """Read platform numbers from $PYORBITAL_CONFIG_PATH/platforms.txt."""
    out_dict = {}

    with open(filename, "r") as fid:
        for row in fid:
            # skip comment lines
            if not row.startswith("#"):
                parts = row.split()
                if len(parts) < 2:
                    continue
                # The satellite name might have whitespace
                platform = " ".join(parts[:-1])
                num = parts[-1]
                if in_upper:
                    platform = platform.upper()
                if num_as_int:
                    num = int(num)
                out_dict[platform] = num

    return out_dict


SATELLITES = read_platform_numbers(get_platforms_filepath(),
                                   in_upper=True, num_as_int=False)
"""
The platform numbers are given in a file $PYORBITAL_CONFIG_PATH/platforms.txt
in the following format:

.. literalinclude:: ../../pyorbital/etc/platforms.txt
  :language: text
  :lines: 5-
"""


def check_is_platform_supported(satname):
    """Check if satellite is supported and print info."""
    if satname in SATELLITES:
        LOGGER.info("Satellite {name} is supported. NORAD number: {norad}".format(
            name=satname, norad=SATELLITES[satname]))
    else:
        LOGGER.info("Satellite {name} is NOT supported.".format(name=satname))
        LOGGER.info("Please add it to a local copy of the platforms.txt file and put in " +
                    "the directory pointed to by the environment variable PYORBITAL_CONFIG_PATH")

    LOGGER.info("Satellite names and NORAD numbers are defined in {filepath}".format(
        filepath=get_platforms_filepath()))


def _dummy_open_stringio(stream):
    return stream


def read(platform, tle_file=None, line1=None, line2=None):
    """Read TLE for *platform*.

    The data are read from *tle_file*, from *line1* and *line2*, from
    the newest file provided in the TLES pattern, or from internet if
    none is provided.

    """
    return Tle(platform, tle_file=tle_file, line1=line1, line2=line2)


def fetch(destination):
    """Fetch TLE from internet and save it to `destination`."""
    with io.open(destination, mode="w", encoding="utf-8") as dest:
        for url in TLE_URLS:
            if not url.lower().startswith("http"):
                raise ValueError(f"{str(url)} is not accepted!")
            response = urlopen(url)  # nosec
            dest.write(response.read().decode("utf-8"))


class ChecksumError(Exception):
    """ChecksumError."""


class Tle:
    """Class holding TLE objects."""

    def __init__(self, platform, tle_file=None, line1=None, line2=None):
        """Init."""
        self._platform = platform.strip().upper()
        self._tle_file = tle_file
        self._line1 = line1
        self._line2 = line2

        self.satnumber = None
        self.classification = None
        self.id_launch_year = None
        self.id_launch_number = None
        self.id_launch_piece = None
        self.epoch_year = None
        self.epoch_day = None
        self.epoch = None
        self.mean_motion_derivative = None
        self.mean_motion_sec_derivative = None
        self.bstar = None
        self.ephemeris_type = None
        self.element_number = None
        self.inclination = None
        self.right_ascension = None
        self.eccentricity = None
        self.arg_perigee = None
        self.mean_anomaly = None
        self.mean_motion = None
        self.orbit = None

        self._read_tle()
        self._checksum()
        self._parse_tle()

    @property
    def excentricity(self):
        """Get 'eccentricity' using legacy 'excentricity' name."""
        warnings.warn("The 'eccentricity' property is deprecated in favor of 'eccentricity'", stacklevel=2)
        return self.eccentricity

    @property
    def line1(self):
        """Return first TLE line."""
        return self._line1

    @property
    def line2(self):
        """Return second TLE line."""
        return self._line2

    @property
    def platform(self):
        """Return satellite platform name."""
        return self._platform

    def _checksum(self):
        """Calculate checksum for the current TLE."""
        for line in [self._line1, self._line2]:
            check = 0
            for char in line[:-1]:
                if char.isdigit():
                    check += int(char)
                if char == "-":
                    check += 1

            if (check % 10) != int(line[-1]):
                raise ChecksumError(self._platform + " " + line)

    def _read_tle(self):
        """Read TLE data."""
        if self._line1 is not None and self._line2 is not None:
            tle = self._line1.strip() + "\n" + self._line2.strip()
        else:
            uris, open_func = _get_uris_and_open_func(tle_file=self._tle_file)
            tle = _get_first_tle(uris, open_func, platform=self._platform)

            if not tle:
                raise KeyError("Found no TLE entry for '%s'" % self._platform)

        self._line1, self._line2 = tle.split("\n")

    def _parse_tle(self):
        """Parse values from TLE data."""
        def _read_tle_decimal(rep):
            """Convert *rep* to decimal value."""
            if rep[0] in ["-", " ", "+"]:
                digits = rep[1:-2].strip()
                val = rep[0] + "." + digits + "e" + rep[-2:]
            else:
                digits = rep[:-2].strip()
                val = "." + digits + "e" + rep[-2:]

            return float(val)

        self.satnumber = self._line1[2:7]
        self.classification = self._line1[7]
        self.id_launch_year = self._line1[9:11]
        self.id_launch_number = self._line1[11:14]
        self.id_launch_piece = self._line1[14:17]
        self.epoch_year = self._line1[18:20]
        self.epoch_day = float(self._line1[20:32])
        self.epoch = \
            np.datetime64(dt.datetime.strptime(self.epoch_year, "%y") +
                          dt.timedelta(days=self.epoch_day - 1), "us")
        self.mean_motion_derivative = float(self._line1[33:43])
        self.mean_motion_sec_derivative = _read_tle_decimal(self._line1[44:52])
        self.bstar = _read_tle_decimal(self._line1[53:61])
        try:
            self.ephemeris_type = int(self._line1[62])
        except ValueError:
            self.ephemeris_type = 0
        self.element_number = int(self._line1[64:68])

        self.inclination = float(self._line2[8:16])
        self.right_ascension = float(self._line2[17:25])
        self.eccentricity = int(self._line2[26:33]) * 10 ** -7
        self.arg_perigee = float(self._line2[34:42])
        self.mean_anomaly = float(self._line2[43:51])
        self.mean_motion = float(self._line2[52:63])
        self.orbit = int(self._line2[63:68])

    def to_dict(self):
        """Return the raw, parsed TLE elements as a dictionary."""
        return {
            "platform": self.platform,
            "satnumber": self.satnumber,
            "classification": self.classification,
            "id_launch_year": self.id_launch_year,
            "id_launch_number": self.id_launch_number,
            "id_launch_piece": self.id_launch_piece,
            "epoch_year": self.epoch_year,
            "epoch_day": self.epoch_day,
            "epoch": self.epoch,
            "mean_motion_derivative": self.mean_motion_derivative,
            "mean_motion_sec_derivative": self.mean_motion_sec_derivative,
            "bstar": self.bstar,
            "ephemeris_type": self.ephemeris_type,
            "element_number": self.element_number,
            "inclination": self.inclination,
            "right_ascension": self.right_ascension,
            "eccentricity": self.eccentricity,
            "arg_perigee": self.arg_perigee,
            "mean_anomaly": self.mean_anomaly,
            "mean_motion": self.mean_motion,
            "orbit": self.orbit,
            "line1": self._line1,
            "line2": self._line2,
        }

    def __str__(self):
        """Format the class data for printing."""
        import pprint
        s_var = io.StringIO()
        d_var = dict(([(k, v) for k, v in
                       list(self.__dict__.items()) if k[0] != "_"]))
        pprint.pprint(d_var, s_var)
        return s_var.getvalue()[:-1]


def _get_local_tle_path_from_env():
    """Get the path to possible local TLE files using the environment variable."""
    return os.environ.get("TLES")


def _get_uris_and_open_func(tle_file=None):
    """Get the uri's and the adequate file open call for the TLE files."""
    local_tle_path = _get_local_tle_path_from_env()

    if tle_file:
        uris, open_func = _get_tle_file_uris_and_open_method(tle_file)
    elif local_tle_path:
        uris, open_func = _get_local_uris_and_open_method(local_tle_path)
    else:
        uris, open_func = _get_internet_uris_and_open_method()
    return uris, open_func


def _get_tle_file_uris_and_open_method(tle_file):
    if isinstance(tle_file, io.StringIO):
        uris = (tle_file,)
        open_func = _dummy_open_stringio
    elif "ADMIN_MESSAGE" in tle_file:
        uris = (io.StringIO(read_tle_from_mmam_xml_file(tle_file)),)
        open_func = _dummy_open_stringio
    else:
        uris = (tle_file,)
        open_func = _open
    return uris, open_func


def _open(filename):
    return io.open(filename, "rb")


def _get_local_uris_and_open_method(local_tle_path):
    # TODO: get the TLE file closest in time to the actual satellite
    # overpass, NOT the latest!
    list_of_tle_files = glob.glob(local_tle_path)
    if list_of_tle_files:
        uris = (max(list_of_tle_files, key=os.path.getctime), )
        LOGGER.debug("Reading TLE from %s", uris[0])
        open_func = _open
    else:
        if config.get("fetch_from_celestrak", None) is not True:
            warn("In the future, implicit downloads of TLEs from Celestrak will be disabled by default. "
                 "You can enable it (and remove this warning) by setting PYORBITAL_FETCH_FROM_CELESTRAK to True.",
                 DeprecationWarning)
        LOGGER.warning("TLES environment variable points to no TLE files")
        throttle_warning = "TLEs will be downloaded from Celestrak, which can throttle the connection."
        LOGGER.warning(throttle_warning)
        warnings.warn(throttle_warning)

        uris, open_func = _get_internet_uris_and_open_method()

    return uris, open_func


def _get_internet_uris_and_open_method():
    LOGGER.debug("Fetch TLE from the internet.")
    return TLE_URLS, urlopen


def _get_first_tle(uris, open_func, platform=""):
    return _get_tles_from_uris(uris, open_func, platform=platform, only_first=True)


def _get_tles_from_uris(uris, open_func, platform="", only_first=True):
    tles = []
    for url in uris:
        tles += _get_tles_from_url(url, open_func, platform, only_first)
    if only_first:
        if tles:
            return tles[0]
        return ""
    return tles


@contextlib.contextmanager
def _uri_open(uri, open_func):
    file_obj = open_func(uri)
    try:
        yield file_obj
    finally:
        if hasattr(file_obj, "close"):
            file_obj.close()


def _get_tles_from_url(url, open_func, platform, only_first):
    with _uri_open(url, open_func) as fid:
        open_is_dummy = open_func == _dummy_open_stringio
        tles = []
        for l_0 in fid:
            tle = _decode_lines(fid, l_0, platform, only_first, open_is_dummy=open_is_dummy)
            if tle:
                if only_first:
                    return [tle]
                tles.append(tle)
        return tles


def _decode(itm):
    if isinstance(itm, str):
        return itm
    return itm.decode("utf-8")


def _decode_lines(fid, l_0, platform, only_first, open_is_dummy=False):
    designator = "1 " + SATELLITES.get(platform, "")
    tle = ""
    l_0 = _decode(l_0)
    if l_0.strip() == platform and platform != "":
        tle = _decode_lines_with_platform_header(fid)
    elif l_0.strip().startswith(designator):
        tle = _decode_lines_without_platform_header(fid, l_0, platform, only_first, open_is_dummy)
    elif l_0.startswith(platform) and platform not in SATELLITES:
        LOGGER.debug("Found a possible match: %s?", str(l_0.strip()))

    return tle


def _decode_lines_with_platform_header(fid):
    l_1 = _decode(next(fid))
    l_2 = _decode(next(fid))
    tle = _merge_tle_from_two_lines(l_1, l_2)
    return tle


def _decode_lines_without_platform_header(fid, l_0, platform, only_first, open_is_dummy):
    tle = ""
    if (platform in SATELLITES or not only_first) or open_is_dummy:
        l_1 = l_0
        l_2 = _decode(next(fid))
        tle = _merge_tle_from_two_lines(l_1, l_2)
        if platform:
            LOGGER.debug("Found platform %s, ID: %s", platform, SATELLITES[platform])
    return tle


def _merge_tle_from_two_lines(l_1, l_2):
    """Merge line1 and line2 to fulle TLE string."""
    return l_1.strip() + "\n" + l_2.strip()


PLATFORM_NAMES_TABLE = "(satid text primary key, platform_name text)"
SATID_TABLE = ("'{}' (epoch date primary key, tle text, insertion_time date,"
               " source text)")
SATID_VALUES = "INSERT INTO '{}' VALUES (?, ?, ?, ?)"
PLATFORM_VALUES = "INSERT INTO platform_names VALUES (?, ?)"
ISO_TIME_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"


class Downloader(object):
    """Class for downloading TLE data."""

    def __init__(self, config):
        """Init."""
        self.config = config

    def fetch_plain_tle(self):
        """Fetch plain text-formated TLE data."""
        tles = {}
        if "fetch_plain_tle" in self.config["downloaders"]:
            sources = self.config["downloaders"]["fetch_plain_tle"]
            for source in sources:
                tles[source] = []
                failures = []
                for uri in sources[source]:
                    try:
                        req = requests.get(uri, timeout=15)  # 15 seconds
                    except requests.exceptions.Timeout:
                        logging.error(f"Failed to make request to {str(uri)} within 15 seconds!")
                        req = None
                    if req and req.status_code == 200:
                        tles[source] += _parse_tles_for_downloader((req.text,), io.StringIO)
                    else:
                        failures.append(uri)
                if len(failures) > 0:
                    logging.error(
                        "Could not fetch TLEs from %s, %d failure(s): [%s]",
                        source, len(failures), ", ".join(failures))
                logging.info("Downloaded %d TLEs from %s", len(tles[source]), source)
        return tles

    def fetch_spacetrack(self):
        """Fetch TLE data from Space-Track."""
        tles = []
        login_url = "https://www.space-track.org/ajaxauth/login"
        download_url = ("https://www.space-track.org/basicspacedata/query/"
                        "class/tle_latest/ORDINAL/1/NORAD_CAT_ID/%s/format/"
                        "tle")
        download_url = download_url % ",".join(
            [str(key) for key in self.config["platforms"]])

        user = self.config["downloaders"]["fetch_spacetrack"]["user"]
        password = self.config["downloaders"]["fetch_spacetrack"]["password"]
        credentials = {"identity": user, "password": password}

        with requests.Session() as session:
            # Login
            req = session.post(login_url, data=credentials)

            if req.status_code != 200:
                logging.error("Could not login to Space-Track")
                return tles

            # Get the data
            req = session.get(download_url)

            if req.status_code == 200:
                tles = _parse_tles_for_downloader((req.text,), io.StringIO)
            else:
                logging.error("Could not retrieve TLEs from Space-Track")

        logging.info("Downloaded %d TLEs from %s", len(tles), "spacetrack")

        return tles

    def read_tle_files(self):
        """Read TLE data from files."""
        paths = self.config["downloaders"]["read_tle_files"]["paths"]

        # Collect filenames
        fnames = collect_filenames(paths)
        tles = _parse_tles_for_downloader(fnames, open)
        logging.info("Loaded %d TLEs from local files", len(tles))

        return tles

    def read_xml_admin_messages(self):
        """Read Eumetsat admin messages in XML format."""
        paths = self.config["downloaders"]["read_xml_admin_messages"]["paths"]
        tles = read_tles_from_mmam_xml_files(paths)
        logging.info("Loaded %d TLEs from admin message XML files", len(tles))

        return tles


def _parse_tles_for_downloader(item, open_func):
    return [Tle("", tle_file=io.StringIO(tle)) for tle in
            _get_tles_from_uris(item, open_func, platform="", only_first=False)]


def collect_filenames(paths):
    """Collect all filenames from *paths*."""
    fnames = []
    for path in paths:
        if "*" in path:
            fnames += glob.glob(path)
        else:
            if not os.path.exists(path):
                logging.error("File %s doesn't exist.", path)
                continue
            fnames += [path]
    return fnames


def read_tles_from_mmam_xml_files(paths):
    """Read TLE data from a list of MMAM XMl file (EUMETSAT).

    MMAM = Multi-Mission Administration Message
    """
    fnames = collect_filenames(paths)
    tles = []
    for fname in fnames:
        data = read_tle_from_mmam_xml_file(fname).split("\n")
        for two_lines in _group_iterable_to_chunks(2, data):
            tl_stream = io.StringIO("\n".join(two_lines))
            tles.append(Tle("", tle_file=tl_stream))
    return tles


def read_tle_from_mmam_xml_file(fname):
    """Read TLE data from MMAM XMl file (EUMETSAT)."""
    tree = ET.parse(fname)
    root = tree.getroot()
    data = []
    for nav in root.findall(".//navigation"):
        data.append(nav.find(".//line-1").text)
        data.append(nav.find(".//line-2").text)

    return "\n".join(data)


def _group_iterable_to_chunks(n, iterable, fillvalue=None):
    """Collect data into fixed-length chunks or blocks."""
    # _group_iterable_to_chunks(3, 'ABCDEFG', 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return zip_longest(fillvalue=fillvalue, *args)


class SQLiteTLE(object):
    """Store TLE data in a sqlite3 database."""

    def __init__(self, db_location, platforms, writer_config):
        """Init."""
        self.db = sqlite3.connect(db_location)
        self.platforms = platforms
        self.writer_config = writer_config
        self.updated = False

        # Create platform_names table if it doesn't exist
        if not table_exists(self.db, "platform_names"):
            cmd = "CREATE TABLE platform_names " + PLATFORM_NAMES_TABLE
            with self.db:
                self.db.execute(cmd)
                logging.info("Created database table 'platform_names'")

    def update_db(self, tle, source):
        """Update the collected data.

        Only data with newer epoch than the existing one is used.

        """
        num = int(tle.satnumber)
        if num not in self.platforms:
            return
        tle.platform_name = self.platforms[num]
        if not table_exists(self.db, num):
            cmd = "CREATE TABLE " + SATID_TABLE.format(num)
            with self.db:
                self.db.execute(cmd)
                logging.info("Created database table '%d'", num)
            cmd = ""
            with self.db:
                self.db.execute(PLATFORM_VALUES, (num, self.platforms[num]))
                logging.info("Added platform name '%s' for ID '%d'",
                             self.platforms[num], num)
        cmd = SATID_VALUES.format(num)
        epoch = tle.epoch.item().isoformat()
        tle = "\n".join([tle.line1, tle.line2])
        now = _utcnow().isoformat()
        try:
            with self.db:
                self.db.execute(cmd, (epoch, tle, now, source))
                logging.info("Added TLE for %d (%s), epoch: %s, source: %s",
                             num, self.platforms[num], epoch, source)
                self.updated = True
        except sqlite3.IntegrityError:
            pass

    def write_tle_txt(self):
        """Write TLE data to a text file."""
        if not self.updated and not self.writer_config.get("write_always",
                                                           False):
            return
        pattern = os.path.join(self.writer_config["output_dir"],
                               self.writer_config["filename_pattern"])
        now = _utcnow()
        fname = now.strftime(pattern)
        out_dir = os.path.dirname(fname)
        if not os.path.exists(out_dir):
            os.makedirs(out_dir)
            logging.info("Created directory %s", out_dir)
        data = []

        for satid, platform_name in self.platforms.items():
            if self.writer_config.get("write_name", False):
                data.append(platform_name)
            query = f"SELECT epoch, tle FROM '{satid:d}' ORDER BY epoch DESC LIMIT 1"  # nosec
            epoch, tle = self.db.execute(query).fetchone()  # nosec
            date_epoch = dt.datetime.strptime(epoch, ISO_TIME_FORMAT)
            tle_age = (_utcnow() - date_epoch).total_seconds() / 3600.
            logging.info("Latest TLE for '%s' (%s) is %d hours old.",
                         satid, platform_name, int(tle_age))
            data.append(tle)

        with open(fname, "w") as fid:
            fid.write("\n".join(data))
            # Add a line-change after the last entry
            fid.write("\n")

        logging.info("Wrote %d TLEs to %s", len(data), fname)

    def close(self):
        """Close the database."""
        self.db.close()


def table_exists(db, name):
    """Check if the table 'name' exists in the database."""
    name = str(name)
    query = "SELECT 1 FROM sqlite_master WHERE type='table' and name=?"
    return db.execute(query, (name,)).fetchone() is not None  # nosec


def _utcnow():
    return dt.datetime.now(tz=dt.timezone.utc).replace(tzinfo=None)

def main():
    """Run a test TLE reading."""
    tle_data = read("Noaa-19")
    print(tle_data)


if __name__ == "__main__":
    main()



"""Unit testing the Astronomy methods and functions."""


import datetime as dt

import dask.array as da
import numpy as np
import numpy.typing as npt
import pytest

import pyorbital.astronomy as astr

try:
    from xarray import DataArray
except ImportError:
    DataArray = None


def _create_dask_array(input_list: list, dtype: npt.DTypeLike) -> da.Array:
    """Create a dummy dask array for testing."""
    np_arr = np.array(input_list, dtype=dtype)
    return da.from_array(np_arr)


def _create_xarray_numpy(input_list: list, dtype: npt.DTypeLike) -> DataArray:
    """Create a dummy xarray DataArray for testing."""
    np_arr = np.array(input_list, dtype=dtype)
    return DataArray(np_arr)


def _create_xarray_dask(input_list: list, dtype: npt.DTypeLike) -> DataArray:
    """Create a dummy daskified xarray DataArray for testing."""
    dask_arr = _create_dask_array(input_list, dtype)
    return DataArray(dask_arr)


class TestAstronomy:
    """Testing the Astronomy class."""

    @pytest.mark.parametrize(
        ("dat", "exp_jdays", "exp_j2000"),
        [
            (dt.datetime(2000, 1, 1, 12, 0), 2451545.0, 0),
            (dt.datetime(2009, 10, 8, 14, 30), 2455113.1041666665, 3568.1041666666665),
        ]
    )
    def test_jdays(self, dat, exp_jdays, exp_j2000):
        """Test julian day functions."""
        assert astr.jdays(dat) == exp_jdays
        assert astr.jdays2000(dat) == exp_j2000

    @pytest.mark.parametrize(
        ("lon", "lat", "exp_theta"),
        [
            # Norrkoping
            (16.1833, 58.6167, 60.371433482557833),
            (0.0, 0.0, 1.8751916863323426),
        ]
    )
    @pytest.mark.parametrize(
        ("dtype", "array_construct"),
        [
            (None, None),
            (np.float32, np.array),
            (np.float64, np.array),
            (np.float32, _create_dask_array),
            (np.float64, _create_dask_array),
            (np.float32, _create_xarray_numpy),
            (np.float64, _create_xarray_numpy),
            (np.float32, _create_xarray_dask),
            (np.float64, _create_xarray_dask),
        ]
    )
    def test_sunangles(self, lon, lat, exp_theta, dtype, array_construct):
        """Test the sun-angle calculations."""
        if array_construct is None and dtype is not None:
            pytest.skip(reason="Xarray dependency unavailable")

        time_slot = dt.datetime(2011, 9, 23, 12, 0)
        abs_tolerance = 1e-8
        if dtype is not None:
            lon = array_construct([lon], dtype=dtype)
            lat = array_construct([lat], dtype=dtype)
            if np.dtype(dtype).itemsize < 8:
                abs_tolerance = 1e-4

        sun_theta = astr.sun_zenith_angle(time_slot, lon, lat)
        if dtype is None:
            assert sun_theta == pytest.approx(exp_theta, abs=abs_tolerance)
            assert isinstance(sun_theta, float)
        else:
            assert sun_theta.dtype == dtype
            np.testing.assert_allclose(sun_theta, exp_theta, atol=abs_tolerance)
            assert isinstance(sun_theta, type(lon))

    def test_sun_earth_distance_correction(self):
        """Test the sun-earth distance correction."""
        utc_time = dt.datetime(2022, 6, 15, 12, 0, 0)
        corr = astr.sun_earth_distance_correction(utc_time)
        corr_exp = 1.0156952156742332
        assert corr == pytest.approx(corr_exp, abs=1e-8)



"""Check if a satellite is supported on default.

If not the name and its NORAD number needs to be added to a local copy of the
platforms.txt file, which then needs to be placed in the directory pointed to
by the environment variable PYORBITAL_CONFIG_PATH.

"""

import argparse
import logging

from pyorbital.logger import logging_on
from pyorbital.tlefile import check_is_platform_supported

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Check if a satellite is supported.")
    parser.add_argument("-s", "--satellite",
                        help=("Name of the Satellite [in upper case] - following WMO Oscar naming."),
                        default=None,
                        required=True,
                        type=str)

    args = parser.parse_args()
    satellite_name = args.satellite

    logging_on(logging.INFO)
    check_is_platform_supported(satellite_name)



"""Test the orbit orbit elements."""

import datetime

import numpy as np
import pytest

from pyorbital.orbital import (
    AE,
    ECC_EPS,
    ECC_LIMIT_HIGH,
    XKMPER,
    XMNPDA,
    OrbitElements,
)


class MockTLE:
    """Mock TLE object for testing OrbitElements."""

    def __init__(self):
        """Initialize mock TLE values for testing."""
        self.epoch = datetime.datetime(2025, 9, 30, 12, 0, 0)
        self.eccentricity = 0.001
        self.inclination = 98.7
        self.right_ascension = 120.0
        self.arg_perigee = 87.0
        self.mean_anomaly = 0.0
        self.mean_motion = 14.2
        self.mean_motion_derivative = 0.0
        self.mean_motion_sec_derivative = 0.0
        self.bstar = 0.0001


def test_orbit_elements_computation():
    """Test basic orbital element computations (SMA, Period, Angle Range)."""
    tle = MockTLE()
    orbit = OrbitElements(tle)

    assert isinstance(orbit.semi_major_axis, float)
    assert orbit.semi_major_axis > 0
    assert orbit.period > 0
    assert -np.pi <= orbit.right_ascension_lon <= np.pi


def test_zero_eccentricity():
    """Test perigee calculation with zero eccentricity."""
    tle = MockTLE()
    tle.eccentricity = 0.0
    orbit = OrbitElements(tle)
    # Perigee == Apogee altitude when e=0
    assert orbit.perigee == pytest.approx(
        (orbit.semi_major_axis / AE - AE) * XKMPER, abs=1e-3
    )


def test_retrograde_orbit_inclination():
    """Test basic inclination check for retrograde orbit."""
    tle = MockTLE()
    tle.inclination = 120.0
    orbit = OrbitElements(tle)
    assert orbit.inclination > np.pi / 2


def test_ra_lon_wrapping():
    """Test longitude wrapping of right ascension with input > 360 deg."""
    tle = MockTLE()
    tle.right_ascension = 400.0  # degrees, > 360
    orbit = OrbitElements(tle)
    assert -np.pi <= orbit.right_ascension_lon <= np.pi


def test_mean_motion_conversion():
    """Test conversion of mean motion to radians per minute."""
    tle = MockTLE()
    orbit = OrbitElements(tle)
    expected = tle.mean_motion * (2 * np.pi / XMNPDA)
    assert orbit.mean_motion == pytest.approx(expected, rel=1e-6)


@pytest.mark.parametrize("incl_deg", [0, 45, 90, 135])
def test_semi_major_axis_vs_inclination(incl_deg):
    """Test semi-major axis computation across inclinations."""
    tle = MockTLE()
    tle.inclination = incl_deg
    orbit = OrbitElements(tle)
    assert orbit.semi_major_axis > 0


@pytest.mark.parametrize("bstar", [0.0, 1e-5, 0.0001, 0.01])
def test_bstar_scaling(bstar):
    """Test scaling of bstar drag term."""
    tle = MockTLE()
    tle.bstar = bstar
    orbit = OrbitElements(tle)
    assert orbit.bstar == pytest.approx(bstar * AE)


def test_mean_motion_derivatives():
    """Test mean motion derivatives are correctly computed."""
    tle = MockTLE()
    orbit = OrbitElements(tle)
    tle.mean_motion_derivative = 1e-5
    tle.mean_motion_sec_derivative = 1e-7
    orbit = OrbitElements(tle)
    assert orbit.mean_motion_derivative > 0
    assert orbit.mean_motion_sec_derivative > 0


def test_apogee_computation():
    """Test apogee altitude calculation."""
    tle = MockTLE()
    orbit = OrbitElements(tle)
    expected_apogee = (
        (orbit.semi_major_axis * (1 + orbit.eccentricity)) / AE - AE
    ) * XKMPER
    assert orbit.apogee == pytest.approx(expected_apogee, abs=1e-3)


def test_semi_major_axis_accuracy():
    """Test semi-major axis against analytical reference value (in km)."""
    tle = MockTLE()
    orbit = OrbitElements(tle)
    expected_sma_km = 7200.645659667062
    computed_sma_km = orbit.semi_major_axis * XKMPER / AE
    assert computed_sma_km == pytest.approx(expected_sma_km, abs=0.1)


def test_orbital_period_accuracy():
    """Test orbital period against corrected mean motion (in seconds)."""
    tle = MockTLE()
    orbit = OrbitElements(tle)
    expected_period_sec = 6080.901176943267
    computed_period_sec = orbit.period * 60
    assert computed_period_sec == pytest.approx(expected_period_sec, abs=1.0)


def test_velocity_at_perigee_accuracy():
    """Test orbital velocity at perigee against analytical reference."""
    tle = MockTLE()
    orbit = OrbitElements(tle)
    expected_velocity_kms = 7.44762253625217
    computed_velocity_kms = orbit.velocity_at_perigee()
    assert computed_velocity_kms == pytest.approx(expected_velocity_kms, abs=0.005)


def test_velocity_at_apogee_accuracy():
    """Test orbital velocity at apogee against analytical reference."""
    tle = MockTLE()
    orbit = OrbitElements(tle)
    expected_velocity_kms = 7.432742171544375
    computed_velocity_kms = orbit.velocity_at_apogee()
    assert computed_velocity_kms == pytest.approx(expected_velocity_kms, abs=0.005)


def test_position_vector_in_orbital_plane_perigee_accuracy():
    """Test position vector at perigee (Mean Anomaly = 0°)."""
    tle = MockTLE()
    tle.eccentricity = 0.001
    tle.mean_anomaly = 0.0
    orbit = OrbitElements(tle)
    pos = orbit.position_vector_in_orbital_plane()
    expected_x = 1.1278289051591721  # Earth radii
    expected_y = 0.0
    assert pos[0] == pytest.approx(expected_x, rel=1e-6)
    assert pos[1] == pytest.approx(expected_y, abs=1e-8)


@pytest.mark.parametrize(
    ("mean_anomaly_deg", "expected_radius"),
    [
        (0, 7193.445014007397),
        (90, 7200.6528603079205),
        (180, 7207.84630532673),
        (270, 7200.6528603079205),
    ],
)
def test_position_vector_in_orbital_plane_varied(mean_anomaly_deg, expected_radius):
    """Verify position vector magnitude matches expected radius at given mean anomaly."""
    tle = MockTLE()
    tle.mean_anomaly = mean_anomaly_deg
    orbit = OrbitElements(tle)
    pos = orbit.position_vector_in_orbital_plane()
    actual_radius = np.linalg.norm(pos) * XKMPER
    assert np.isclose(actual_radius, expected_radius, rtol=1e-6)


@pytest.mark.parametrize(
    ("eccentricity", "expected"),
    [
        (0.0005, True),
        (0.01, False),
    ],
)
def test_is_circular_property(eccentricity, expected):
    """Test circular orbit detection based on eccentricity."""
    tle = MockTLE()
    tle.eccentricity = eccentricity
    orbit = OrbitElements(tle)
    assert orbit.is_circular == expected


@pytest.mark.parametrize(
    ("inclination_deg", "expected"),
    [
        (100.0, True),
        (80.0, False),
    ],
)
def test_is_retrograde_property(inclination_deg, expected):
    """Test retrograde orbit detection based on inclination."""
    tle = MockTLE()
    tle.inclination = inclination_deg
    orbit = OrbitElements(tle)
    assert orbit.is_retrograde == expected


@pytest.mark.parametrize("eccentricity", [0.001, 0.01, 0.1, 0.5])
def test_velocity_perigee_greater_than_apogee(eccentricity):
    """Ensure velocity at perigee is greater than at apogee for elliptical orbits."""
    tle = MockTLE()
    tle.eccentricity = eccentricity
    orbit = OrbitElements(tle)
    v_perigee = orbit.velocity_at_perigee()
    v_apogee = orbit.velocity_at_apogee()
    assert v_perigee > v_apogee


@pytest.mark.parametrize("eccentricity", [0.0, ECC_EPS / 10, ECC_EPS])
def test_velocity_equal_for_circular_orbits(eccentricity):
    """Ensure velocity at perigee equals velocity at apogee for nearly circular orbits."""
    tle = MockTLE()
    tle.eccentricity = eccentricity
    orbit = OrbitElements(tle)
    v_perigee = orbit.velocity_at_perigee()
    v_apogee = orbit.velocity_at_apogee()
    assert v_perigee == pytest.approx(v_apogee, rel=1e-5)


def test_high_eccentricity_limit():
    """Test behavior near the upper eccentricity limit."""
    tle = MockTLE()
    tle.eccentricity = ECC_LIMIT_HIGH
    orbit = OrbitElements(tle)
    assert orbit.semi_major_axis > 0
    assert orbit.velocity_at_perigee() > orbit.velocity_at_apogee()


def test_sgp4_unnormalization_value():
    """Validate SGP4 un-normalization of mean motion and semi-major axis.

    This test uses hardcoded reference values from validated SGP4 output.
    """

    class RefTLE(MockTLE):
        """Reference TLE with fixed orbital parameters for precision testing."""

        def __init__(self):
            super().__init__()
            """Initialize reference TLE values for SGP4 unnormalization test."""
            self.inclination = 98.7408  # degrees
            self.eccentricity = 0.001140
            self.mean_motion = 14.28634842  # rev/day
            self.epoch = datetime.datetime(2200, 1, 1, 0, 0, 0)

    tle = RefTLE()
    orbit = OrbitElements(tle)

    # Hardcoded expected values from validated SGP4 propagation
    expected_mean_motion = 0.06237319306246916  # rad/min
    expected_semi_major_axis = 1.1244009310620886  # Earth radii

    assert orbit.original_mean_motion == pytest.approx(expected_mean_motion, rel=1e-8)
    assert orbit.semi_major_axis == pytest.approx(expected_semi_major_axis, rel=1e-8)



"""Test the geoloc orbital."""

import datetime as dt
import unittest
from unittest import mock

import numpy as np
import pytest

from pyorbital import orbital

eps_deg = 10e-3


class Test(unittest.TestCase):
    """Basic test class for unittesting the pyorbital.orbital class."""

    def test_get_orbit_number(self):
        """Testing getting the orbitnumber from the TLEs."""
        sat = orbital.Orbital("NPP",
                              line1="1 37849U 11061A   12017.90990040 "
                                    "-.00000112  00000-0 -32693-4 0   772",
                              line2="2 37849  98.7026 317.8811 0001845  "
                                    "92.4533 267.6830 14.19582686 11574")
        dobj = dt.datetime(2012, 1, 18, 8, 4, 19)
        orbnum = sat.get_orbit_number(dobj)
        assert orbnum == 1163

    def test_sublonlat(self):
        """Test getting the sub-satellite position."""
        sat = orbital.Orbital("ISS (ZARYA)",
                              line1="1 25544U 98067A   03097.78853147  "
                                    ".00021906  00000-0  28403-3 0  8652",
                              line2="2 25544  51.6361  13.7980 0004256  "
                                    "35.6671  59.2566 15.58778559250029")
        d = dt.datetime(2003, 3, 23, 0, 3, 22)
        lon, lat, alt = sat.get_lonlatalt(d)
        expected_lon = -68.199894472013213
        expected_lat = 23.159747677881075
        expected_alt = 392.01953430856935
        assert np.abs(lon - expected_lon) < eps_deg, "Calculation of sublon failed"
        assert np.abs(lat - expected_lat) < eps_deg, "Calculation of sublat failed"
        assert np.abs(alt - expected_alt) < eps_deg, "Calculation of altitude failed"

    def test_observer_look(self):
        """Test getting the observer look angles."""
        sat = orbital.Orbital("ISS (ZARYA)",
                              line1="1 25544U 98067A   03097.78853147  "
                                    ".00021906  00000-0  28403-3 0  8652",
                              line2="2 25544  51.6361  13.7980 0004256  "
                                    "35.6671  59.2566 15.58778559250029")
        d = dt.datetime(2003, 3, 23, 0, 3, 22)
        az, el = sat.get_observer_look(d, -84.39733, 33.775867, 0)
        expected_az = 122.45169655331965
        expected_el = 1.9800219611255456
        assert np.abs(az - expected_az) < eps_deg, "Calculation of azimut failed"
        assert np.abs(el - expected_el) < eps_deg, "Calculation of elevation failed"

    def test_orbit_num_an(self):
        """Test getting orbit number - ascending node."""
        sat = orbital.Orbital("METOP-A",
                              line1="1 29499U 06044A   11254.96536486  "
                                    ".00000092  00000-0  62081-4 0  5221",
                              line2="2 29499  98.6804 312.6735 0001758 "
                                    "111.9178 248.2152 14.21501774254058")
        d = dt.datetime(2011, 9, 14, 5, 30)
        assert sat.get_orbit_number(d) == 25437

    def test_orbit_num_non_an(self):
        """Test getting orbit number - not ascending node."""
        sat = orbital.Orbital("METOP-A",
                              line1="1 29499U 06044A   13060.48822809  "
                                    ".00000017  00000-0  27793-4 0  9819",
                              line2="2 29499  98.6639 121.6164 0001449  "
                                    "71.9056  43.3132 14.21510544330271")
        dt = np.timedelta64(98, "m")
        assert sat.get_orbit_number(sat.tle.epoch + dt) == 33028

    def test_orbit_num_equator(self):
        """Test getting orbit numbers when being around equator."""
        sat = orbital.Orbital("SUOMI NPP",
                              line1="1 37849U 11061A   13061.24611272  "
                                    ".00000048  00000-0  43679-4 0  4334",
                              line2="2 37849  98.7444   1.0588 0001264  "
                                    "63.8791 102.8546 14.19528338 69643")
        t1 = dt.datetime(2013, 3, 2, 22, 2, 25)
        t2 = dt.datetime(2013, 3, 2, 22, 2, 26)
        on1 = sat.get_orbit_number(t1)
        on2 = sat.get_orbit_number(t2)
        assert on1 == 6973
        assert on2 == 6974
        pos1, vel1 = sat.get_position(t1, normalize=False)
        pos2, vel2 = sat.get_position(t2, normalize=False)
        del vel1, vel2
        assert pos1[2] < 0
        assert pos2[2] > 0

    def test_get_next_passes_apogee(self):
        """Regression test #22."""
        line1 = "1 24793U 97020B   18065.48735489  " \
                ".00000075  00000-0  19863-4 0  9994"
        line2 = "2 24793  86.3994 209.3241 0002020  " \
                "89.8714 270.2713 14.34246429 90794"

        orb = orbital.Orbital("IRIDIUM 7 [+]", line1=line1, line2=line2)
        d = dt.datetime(2018, 3, 7, 3, 30, 15)
        res = orb.get_next_passes(d, 1, 170.556, -43.368, 0.5, horizon=40)
        assert abs(res[0][2] - dt.datetime(2018, 3, 7, 3, 48, 13, 178439)) < dt.timedelta(seconds=0.01)

    def test_get_next_passes_tricky(self):
        """Check issue #34 for reference."""
        line1 = "1 43125U 18004Q   18251.42128650 " \
            "+.00001666 +00000-0 +73564-4 0  9991"

        line2 = "2 43125 097.5269 314.3317 0010735 "\
            "157.6344 202.5362 15.23132245036381"

        orb = orbital.Orbital("LEMUR-2-BROWNCOW", line1=line1, line2=line2)
        d = dt.datetime(2018, 9, 8)

        res = orb.get_next_passes(d, 72, -8.174163, 51.953319, 0.05, horizon=5)

        assert abs(res[0][2] - dt.datetime(2018, 9, 8, 9, 5, 46, 375248)) < dt.timedelta(seconds=0.01)
        assert abs(res[-1][2] - dt.datetime(2018, 9, 10, 22, 15, 3, 143469)) < dt.timedelta(seconds=0.01)

        assert len(res) == 15

    def test_get_next_passes_issue_22(self):
        """Check that max."""
        line1 = "1 28654U 05018A   21083.16603416  .00000102  00000-0  79268-4 0  9999"
        line2 = "2 28654  99.0035 147.6583 0014816 159.4931 200.6838 14.12591533816498"

        orb = orbital.Orbital("NOAA 18", line1=line1, line2=line2)
        t = dt.datetime(2021, 3, 9, 22)
        next_passes = orb.get_next_passes(t, 1, -15.6335, 27.762, 0.)
        rise, fall, max_elevation = next_passes[0]
        assert rise < max_elevation < fall

    @mock.patch("pyorbital.orbital.Orbital.get_lonlatalt")
    def test_utc2local(self, get_lonlatalt):
        """Test converting UTC to local time."""
        get_lonlatalt.return_value = -45, None, None
        sat = orbital.Orbital("METOP-A",
                              line1="1 29499U 06044A   13060.48822809  "
                                    ".00000017  00000-0  27793-4 0  9819",
                              line2="2 29499  98.6639 121.6164 0001449  "
                                    "71.9056  43.3132 14.21510544330271")
        assert sat.utc2local(dt.datetime(2009, 7, 1, 12)) == dt.datetime(2009, 7, 1, 9)

    @mock.patch("pyorbital.orbital.Orbital.utc2local")
    @mock.patch("pyorbital.orbital.Orbital.get_orbit_number")
    def test_get_equatorial_crossing_time(self, get_orbit_number, utc2local):
        """Test get the equatorial crossing time."""
        def get_orbit_number_patched(utc_time, **kwargs):
            utc_time = np.datetime64(utc_time)
            diff = (utc_time - np.datetime64("2009-07-01 12:38:12")) / np.timedelta64(7200, "s")
            return 1234 + diff

        get_orbit_number.side_effect = get_orbit_number_patched
        utc2local.return_value = "local_time"
        sat = orbital.Orbital("METOP-A",
                              line1="1 29499U 06044A   13060.48822809  "
                                    ".00000017  00000-0  27793-4 0  9819",
                              line2="2 29499  98.6639 121.6164 0001449  "
                                    "71.9056  43.3132 14.21510544330271")

        # Ascending node
        res = sat.get_equatorial_crossing_time(tstart=dt.datetime(2009, 7, 1, 12),
                                               tend=dt.datetime(2009, 7, 1, 13))
        exp = dt.datetime(2009, 7, 1, 12, 38, 12)
        assert res - exp < dt.timedelta(seconds=0.01)

        # Descending node
        res = sat.get_equatorial_crossing_time(tstart=dt.datetime(2009, 7, 1, 12),
                                               tend=dt.datetime(2009, 7, 1, 14, 0),
                                               node="descending")
        exp = dt.datetime(2009, 7, 1, 13, 38, 12)
        assert res - exp < dt.timedelta(seconds=0.01)

        # Conversion to local time
        res = sat.get_equatorial_crossing_time(tstart=dt.datetime(2009, 7, 1, 12),
                                               tend=dt.datetime(2009, 7, 1, 14),
                                               local_time=True)
        assert res == "local_time"


class TestGetObserverLook(unittest.TestCase):
    """Test the get_observer_look function."""

    def setUp(self):
        """Set up the test environment."""
        self.t = dt.datetime(2018, 1, 1, 0, 0, 0)
        self.sat_lon = np.array([[-89.5, -89.4, -89.5, -89.4],
                                 [-89.3, -89.2, -89.3, -89.2]])
        self.sat_lat = np.array([[45.5, 45.4, 45.5, 45.4],
                                 [45.3, 40.2, 45.3, 40.2]])
        self.sat_alt = 35786 * np.ones((2, 4))
        self.lon = np.array([[-85.5, -85.4, -89.5, -99.4],
                             [-85.3, -89.2, -89.3, -79.2]])
        self.lat = np.array([[40.5, 40.4, 65.5, 45.4],
                             [40.3, 40.2, 25.3, 40.2]])
        self.alt = np.zeros((2, 4))
        self.exp_azi = np.array([[331.00275902, 330.95954165, 180, 86.435411],
                                 [330.91642994, 180, 0, 273.232073]])
        self.exp_elev = np.array([[83.18070976, 83.17788976, 66.548467, 81.735221],
                                  [83.17507167, 90, 66.559906, 81.010018]])

    def test_basic_numpy(self):
        """Test with numpy array inputs."""
        from pyorbital import orbital
        azi, elev = orbital.get_observer_look(self.sat_lon, self.sat_lat,
                                              self.sat_alt, self.t,
                                              self.lon, self.lat, self.alt)
        np.testing.assert_allclose(azi, self.exp_azi)
        np.testing.assert_allclose(elev, self.exp_elev)

    def test_basic_dask(self):
        """Test with dask array inputs."""
        import dask.array as da

        from pyorbital import orbital
        sat_lon = da.from_array(self.sat_lon, chunks=2)
        sat_lat = da.from_array(self.sat_lat, chunks=2)
        sat_alt = da.from_array(self.sat_alt, chunks=2)
        lon = da.from_array(self.lon, chunks=2)
        lat = da.from_array(self.lat, chunks=2)
        alt = da.from_array(self.alt, chunks=2)
        azi, elev = orbital.get_observer_look(sat_lon, sat_lat,
                                              sat_alt, self.t,
                                              lon, lat, alt)
        np.testing.assert_allclose(azi.compute(), self.exp_azi)
        np.testing.assert_allclose(elev.compute(), self.exp_elev)

    def test_xarray_with_numpy(self):
        """Test with xarray DataArray with numpy array as inputs."""
        import xarray as xr

        from pyorbital import orbital

        def _xarr_conv(input_array):
            return xr.DataArray(input_array)
        sat_lon = _xarr_conv(self.sat_lon)
        sat_lat = _xarr_conv(self.sat_lat)
        sat_alt = _xarr_conv(self.sat_alt)
        lon = _xarr_conv(self.lon)
        lat = _xarr_conv(self.lat)
        alt = _xarr_conv(self.alt)
        azi, elev = orbital.get_observer_look(sat_lon, sat_lat,
                                              sat_alt, self.t,
                                              lon, lat, alt)
        np.testing.assert_allclose(azi.data, self.exp_azi)
        np.testing.assert_allclose(elev.data, self.exp_elev)

    def test_xarray_with_dask(self):
        """Test with xarray DataArray with dask array as inputs."""
        import dask.array as da
        import xarray as xr

        from pyorbital import orbital

        def _xarr_conv(input_array):
            return xr.DataArray(da.from_array(input_array, chunks=2))
        sat_lon = _xarr_conv(self.sat_lon)
        sat_lat = _xarr_conv(self.sat_lat)
        sat_alt = _xarr_conv(self.sat_alt)
        lon = _xarr_conv(self.lon)
        lat = _xarr_conv(self.lat)
        alt = _xarr_conv(self.alt)
        azi, elev = orbital.get_observer_look(sat_lon, sat_lat,
                                              sat_alt, self.t,
                                              lon, lat, alt)
        np.testing.assert_allclose(azi.data.compute(), self.exp_azi)
        np.testing.assert_allclose(elev.data.compute(), self.exp_elev)

    def test_scalar(self):
        """Test with scalar inputs."""
        from pyorbital.orbital import get_observer_look
        (azi, elev) = get_observer_look(0, 0, 30_000_000, self.t, 0, 0, 0)
        np.testing.assert_allclose(elev, 90)


class TestGetObserverLookNadir(unittest.TestCase):
    """Test the get_observer_look function when satellite is at nadir."""

    def setUp(self):
        """Set up for test observer at nadir.

        Note that rounding error differs between array types.
        With 1000 elements a test gives:
          1 error for basic numpy
          41 errors for basic dask
          63 errors for xarray with dask
          2 error for xarray with numpy
        """
        rng = np.random.RandomState(125)
        self.t = dt.datetime(2018, 1, 1, 0, 0, 0)
        self.sat_lon = 360 * rng.rand(100) - 180
        self.sat_lat = 180 * rng.rand(100) - 90
        self.sat_alt = rng.rand(100) + 850
        self.lon = self.sat_lon  # + 10E-17
        self.lat = self.sat_lat  # + 10E-17
        self.alt = np.zeros((100))
        self.exp_elev = np.zeros((100)) + 90

    def test_basic_numpy(self):
        """Test with numpy array inputs."""
        from pyorbital import orbital
        azi, elev = orbital.get_observer_look(self.sat_lon, self.sat_lat,
                                              self.sat_alt, self.t,
                                              self.lon, self.lat, self.alt)
        assert np.sum(np.isnan(azi)) == 0
        assert not np.isnan(azi).any()
        np.testing.assert_allclose(elev, self.exp_elev)

    def test_basic_dask(self):
        """Test with dask array inputs."""
        import dask.array as da

        from pyorbital import orbital
        sat_lon = da.from_array(self.sat_lon, chunks=2)
        sat_lat = da.from_array(self.sat_lat, chunks=2)
        sat_alt = da.from_array(self.sat_alt, chunks=2)
        lon = da.from_array(self.lon, chunks=2)
        lat = da.from_array(self.lat, chunks=2)
        alt = da.from_array(self.alt, chunks=2)
        azi, elev = orbital.get_observer_look(sat_lon, sat_lat,
                                              sat_alt, self.t,
                                              lon, lat, alt)
        assert np.sum(np.isnan(azi)) == 0
        assert not np.isnan(azi).any()
        np.testing.assert_allclose(elev.compute(), self.exp_elev)

    def test_xarray_with_numpy(self):
        """Test with xarray DataArray with numpy array as inputs."""
        import xarray as xr

        from pyorbital import orbital

        def _xarr_conv(input_array):
            return xr.DataArray(input_array)
        sat_lon = _xarr_conv(self.sat_lon)
        sat_lat = _xarr_conv(self.sat_lat)
        sat_alt = _xarr_conv(self.sat_alt)
        lon = _xarr_conv(self.lon)
        lat = _xarr_conv(self.lat)
        alt = _xarr_conv(self.alt)
        azi, elev = orbital.get_observer_look(sat_lon, sat_lat,
                                              sat_alt, self.t,
                                              lon, lat, alt)
        assert np.sum(np.isnan(azi)) == 0
        assert not np.isnan(azi).any()
        np.testing.assert_allclose(elev.data, self.exp_elev)

    def test_xarray_with_dask(self):
        """Test with xarray DataArray with dask array as inputs."""
        import dask.array as da
        import xarray as xr

        from pyorbital import orbital

        def _xarr_conv(input_array):
            return xr.DataArray(da.from_array(input_array, chunks=2))
        sat_lon = _xarr_conv(self.sat_lon)
        sat_lat = _xarr_conv(self.sat_lat)
        sat_alt = _xarr_conv(self.sat_alt)
        lon = _xarr_conv(self.lon)
        lat = _xarr_conv(self.lat)
        alt = _xarr_conv(self.alt)
        azi, elev = orbital.get_observer_look(sat_lon, sat_lat,
                                              sat_alt, self.t,
                                              lon, lat, alt)
        assert np.sum(np.isnan(azi)) == 0
        assert not np.isnan(azi).any()
        np.testing.assert_allclose(elev.data.compute(), self.exp_elev)


class TestRegressions(unittest.TestCase):
    """Test regressions."""

    def test_63(self):
        """Check that no runtimewarning is raised, #63."""
        import warnings

        from pyorbital.orbital import Orbital
        warnings.filterwarnings("error")
        orb = Orbital("Suomi-NPP",
                      line1="1 37849U 11061A   19292.84582509  .00000011  00000-0  25668-4 0  9997",
                      line2="2 37849  98.7092 229.3263 0000715  98.5313 290.6262 14.19554485413345")
        orb.get_next_passes(dt.datetime(2019, 10, 21, 16, 0, 0), 12, 123.29736, -13.93763, 0)
        warnings.filterwarnings("default")


@pytest.mark.parametrize("dtime",
                         [dt.datetime(2024, 6, 25, 11, 0, 18),
                          dt.datetime(2024, 6, 25, 11, 5, 0, 0, dt.timezone.utc),
                          np.datetime64("2024-06-25T11:10:00.000000")
                          ]
                         )
def test_get_last_an_time_scalar_input(dtime):
    """Test getting the time of the last ascending node - input time is a scalar."""
    from pyorbital.orbital import Orbital
    orb = Orbital("NOAA-20",
                  line1="1 43013U 17073A   24176.73674251  .00000000  00000+0  11066-3 0 00014",
                  line2="2 43013  98.7060 114.5340 0001454 139.3958 190.7541 14.19599847341971")

    expected = np.datetime64("2024-06-25T10:44:18.234375")
    result = orb.get_last_an_time(dtime)
    assert abs(expected - result) < np.timedelta64(1, "s")


@pytest.mark.parametrize("dtime",
                         [dt.datetime(2024, 6, 25, 11, 5, 0, 0, dt.timezone(dt.timedelta(hours=1))),
                          ]
                         )
def test_get_last_an_time_wrong_input(dtime):
    """Test getting the time of the last ascending node - wrong input."""
    from pyorbital.orbital import Orbital
    orb = Orbital("NOAA-20",
                  line1="1 43013U 17073A   24176.73674251  .00000000  00000+0  11066-3 0 00014",
                  line2="2 43013  98.7060 114.5340 0001454 139.3958 190.7541 14.19599847341971")

    expected = "UTC time expected! Parsing a timezone aware datetime object requires it to be UTC!"
    with pytest.raises(ValueError, match=expected):
        _ = orb.get_last_an_time(dtime)



"""Test cases from the AIAA article."""


# TODO: right formal unit tests.
from __future__ import print_function, with_statement

import datetime as dt
import os
import unittest

import numpy as np

from pyorbital import astronomy, tlefile
from pyorbital.orbital import _SGDP4, Orbital, OrbitElements
from pyorbital.tlefile import ChecksumError


class LineOrbital(Orbital):
    """Read TLE lines instead of file."""

    def __init__(self, satellite, line1, line2):
        """Initialize the class."""
        satellite = satellite.upper()
        self.satellite_name = satellite
        self.tle = tlefile.read(satellite, line1=line1, line2=line2)
        self.orbit_elements = OrbitElements(self.tle)
        self._sgdp4 = _SGDP4(self.orbit_elements)


def get_results(satnumber, delay):
    """Get expected results from result file."""
    path = os.path.dirname(os.path.abspath(__file__))
    with open(os.path.join(path, "aiaa_results")) as f_2:
        line = f_2.readline()
        while line:
            if line.endswith(" xx\n") and int(line[:-3]) == satnumber:
                line = f_2.readline()
                while (not line.startswith("%.8f" % delay)):
                    line = f_2.readline()
                sline = line.split()
                if delay == 0:
                    utc_time = None
                else:
                    utc_time = dt.datetime.strptime(sline[-1], "%H:%M:%S.%f")
                    utc_time = utc_time.replace(year=int(sline[-4]),
                                                month=int(sline[-3]),
                                                day=int(sline[-2]))
                    utc_time = np.datetime64(utc_time)
                return (float(sline[1]),
                        float(sline[2]),
                        float(sline[3]),
                        float(sline[4]),
                        float(sline[5]),
                        float(sline[6]),
                        utc_time)
            line = f_2.readline()


_DATAPATH = os.path.dirname(os.path.abspath(__file__))


class AIAAIntegrationTest(unittest.TestCase):
    """Test against the AIAA test cases."""

    @unittest.skipIf(
        not os.path.exists(os.path.join(_DATAPATH, "SGP4-VER.TLE")),
        "SGP4-VER.TLE not available")
    def test_aiaa(self):
        """Do the tests against AIAA test cases."""
        path = os.path.dirname(os.path.abspath(__file__))
        with open(os.path.join(path, "SGP4-VER.TLE")) as f__:
            test_line = f__.readline()
            while test_line:
                if test_line.startswith("#"):
                    test_name = test_line
                if test_line.startswith("1 "):
                    line1 = test_line
                if test_line.startswith("2 "):
                    _check_line2(f__, test_name, line1, test_line)

                test_line = f__.readline()


def _check_line2(f__, test_name: str, line1: str, test_line: str) -> None:
    line2 = test_line[:69]
    times_str = str.split(test_line[69:])
    times = np.arange(float(times_str[0]),
                      float(times_str[1]) + 1,
                      float(times_str[2]))
    if test_name.startswith("#   SL-14 DEB"):
        # FIXME: we have to handle decaying satellites!
        return

    try:
        o = LineOrbital("unknown", line1, line2)
    except NotImplementedError:
        return
    except ChecksumError:
        assert test_line.split()[1] in ["33333", "33334", "33335"]
        return

    for delay in times:
        try:
            test_time = delay.astype("timedelta64[m]") + o.tle.epoch
            pos, vel = o.get_position(test_time, False)
            res = get_results(
                int(o.tle.satnumber), float(delay))
        except NotImplementedError:
            # Skipping deep-space
            break
        # except ValueError, e:
        #     from warnings import warn
        #     warn(test_name + ' ' + str(e))
        #     break

        delta_pos = 5e-6  # km =  5 mm
        delta_vel = 5e-9  # km/s = 5 um/s
        delta_time = 1e-3  # 1 millisecond
        assert abs(res[0] - pos[0]) < delta_pos
        assert abs(res[1] - pos[1]) < delta_pos
        assert abs(res[2] - pos[2]) < delta_pos
        assert abs(res[3] - vel[0]) < delta_vel
        assert abs(res[4] - vel[1]) < delta_vel
        assert abs(res[5] - vel[2]) < delta_vel
        if res[6] is not None:
            dt = astronomy._days(res[6] - test_time) * 24 * 60
            assert abs(dt) < delta_time



"""Module for computing the orbital parameters of satellites."""

import datetime as dt
import logging
import warnings
from functools import partial

import numpy as np
from scipy import optimize

from pyorbital import astronomy, dt2np

try:
    import dask.array as da
    has_dask = True
except ImportError:
    da = None
    has_dask = False

try:
    import xarray as xr
    has_xarray = True
except ImportError:
    xr = None
    has_xarray = False

logger = logging.getLogger(__name__)

ECC_EPS = 1.0e-6  # Too low for computing further drops.
ECC_LIMIT_LOW = -1.0e-3
ECC_LIMIT_HIGH = 1.0 - ECC_EPS  # Too close to 1
ECC_ALL = 1.0e-4

EPS_COS = 1.5e-12

NR_EPS = 1.0e-12

CK2 = 5.413080e-4
CK4 = 0.62098875e-6
E6A = 1.0e-6
QOMS2T = 1.88027916e-9
S = 1.01222928
S0 = 78.0
XJ3 = -0.253881e-5
XKE = 0.743669161e-1
XKMPER = 6378.135
XMNPDA = 1440.0
# MFACTOR = 7.292115E-5
AE = 1.0
SECDAY = 8.6400E4

F = 1 / 298.257223563  # Earth flattening WGS-84
A = 6378.137  # WGS84 Equatorial radius


SGDP4_ZERO_ECC = 0
SGDP4_DEEP_NORM = 1
SGDP4_NEAR_SIMP = 2
SGDP4_NEAR_NORM = 3

KS = AE * (1.0 + S0 / XKMPER)
A3OVK2 = (-XJ3 / CK2) * AE**3


class OrbitalError(Exception):
    """Custom exception for the Orbital class."""

    pass


def get_observer_look(sat_lon, sat_lat, sat_alt, utc_time, lon, lat, alt):
    """Calculate observers look angle to a satellite.

    http://celestrak.com/columns/v02n02/

    :utc_time: Observation time (datetime object)
    :lon: Longitude of observer position on ground in degrees east
    :lat: Latitude of observer position on ground in degrees north
    :alt: Altitude above sea-level (geoid) of observer position on ground in km

    :return: (Azimuth, Elevation)
    """
    (pos_x, pos_y, pos_z), (vel_x, vel_y, vel_z) = astronomy.observer_position(
        utc_time, sat_lon, sat_lat, sat_alt)

    (opos_x, opos_y, opos_z), (ovel_x, ovel_y, ovel_z) = \
        astronomy.observer_position(utc_time, lon, lat, alt)

    lon = np.deg2rad(lon)
    lat = np.deg2rad(lat)

    theta = (astronomy.gmst(utc_time) + lon) % (2 * np.pi)

    rx = pos_x - opos_x
    ry = pos_y - opos_y
    rz = pos_z - opos_z

    sin_lat = np.sin(lat)
    cos_lat = np.cos(lat)
    sin_theta = np.sin(theta)
    cos_theta = np.cos(theta)

    top_s = sin_lat * cos_theta * rx + \
        sin_lat * sin_theta * ry - cos_lat * rz
    top_e = -sin_theta * rx + cos_theta * ry
    top_z = cos_lat * cos_theta * rx + \
        cos_lat * sin_theta * ry + sin_lat * rz

    # Azimuth is undefined when elevation is 90 degrees, 180 (pi) will be returned.
    az_ = np.arctan2(-top_e, top_s) + np.pi
    az_ = np.mod(az_, 2 * np.pi)  # Needed on some platforms

    rg_ = np.sqrt(rx * rx + ry * ry + rz * rz)

    top_z_divided_by_rg_ = top_z / rg_

    # Due to rounding top_z can be larger than rg_ (when el_ ~ 90).
    top_z_divided_by_rg_ = top_z_divided_by_rg_.clip(max=1)
    el_ = np.arcsin(top_z_divided_by_rg_)

    return np.rad2deg(az_), np.rad2deg(el_)


class Orbital(object):
    """Class for orbital computations.

    The *satellite* parameter is the name of the satellite to work on and is
    used to retrieve the right TLE data for internet or from *tle_file* in case
    it is provided.
    """

    def __init__(self, satellite, tle_file=None, line1=None, line2=None):
        """Initialize the class."""
        from pyorbital import tlefile

        satellite = satellite.upper()
        self.satellite_name = satellite
        self.tle = tlefile.read(satellite, tle_file=tle_file,
                                line1=line1, line2=line2)
        self.orbit_elements = OrbitElements(self.tle)
        self._sgdp4 = _SGDP4(self.orbit_elements)

    def __str__(self):
        """Print the Orbital object state."""
        return self.satellite_name + " " + str(self.tle)

    def get_last_an_time(self, utc_time):
        """Calculate time of last ascending node relative to the specified time."""
        # Propagate backwards to ascending node
        dt = np.timedelta64(10, "m")

        t_old = np.datetime64(_get_tz_unaware_utctime(utc_time))
        t_new = t_old - dt
        pos0, vel0 = self.get_position(t_old, normalize=False)
        pos1, vel1 = self.get_position(t_new, normalize=False)
        while not (pos0[2] > 0 and pos1[2] < 0):
            pos0 = pos1
            t_old = t_new
            t_new = t_old - dt
            pos1, vel1 = self.get_position(t_new, normalize=False)

        # Return if z within 1 km of an
        if np.abs(pos0[2]) < 1:
            return t_old
        elif np.abs(pos1[2]) < 1:
            return t_new

        # Bisect to z within 1 km
        while np.abs(pos1[2]) > 1:
            # pos0, vel0 = pos1, vel1
            dt = (t_old - t_new) / 2
            t_mid = t_old - dt
            pos1, vel1 = self.get_position(t_mid, normalize=False)
            if pos1[2] > 0:
                t_old = t_mid
            else:
                t_new = t_mid

        return t_mid

    def get_position(self, utc_time, normalize=True):
        """Get the cartesian position and velocity from the satellite."""
        kep = self._sgdp4.propagate(utc_time)
        pos, vel = kep2xyz(kep)

        if normalize:
            pos /= XKMPER
            vel /= XKMPER * XMNPDA / SECDAY

        return pos, vel

    def get_lonlatalt(self, utc_time):
        """Calculate sublon, sublat and altitude of satellite.

        http://celestrak.com/columns/v02n03/
        """
        (pos_x, pos_y, pos_z), (vel_x, vel_y, vel_z) = self.get_position(
            utc_time, normalize=True)

        lon = ((np.arctan2(pos_y * XKMPER, pos_x * XKMPER) - astronomy.gmst(utc_time))
               % (2 * np.pi))

        lon = np.where(lon > np.pi, lon - np.pi * 2, lon)
        lon = np.where(lon <= -np.pi, lon + np.pi * 2, lon)

        r = np.sqrt(pos_x ** 2 + pos_y ** 2)
        lat = np.arctan2(pos_z, r)
        e2 = F * (2 - F)
        while True:
            lat2 = lat
            c = 1 / (np.sqrt(1 - e2 * (np.sin(lat2) ** 2)))
            lat = np.arctan2(pos_z + c * e2 * np.sin(lat2), r)
            if np.all(abs(lat - lat2) < 1e-10):
                break
        alt = r / np.cos(lat) - c
        alt *= A
        return np.rad2deg(lon), np.rad2deg(lat), alt

    def find_aos(self, utc_time, lon, lat):
        """Find AOS."""
        pass

    def find_aol(self, utc_time, lon, lat):
        """Find AOL."""
        pass

    def get_observer_look(self, utc_time, lon, lat, alt):
        """Calculate observers look angle to a satellite.

        See http://celestrak.com/columns/v02n02/

        utc_time: Observation time (datetime object)
        lon: Longitude of observer position on ground in degrees east
        lat: Latitude of observer position on ground in degrees north
        alt: Altitude above sea-level (geoid) of observer position on ground in km

        Return: (Azimuth, Elevation)

        """
        utc_time = dt2np(utc_time)
        (pos_x, pos_y, pos_z), (vel_x, vel_y, vel_z) = self.get_position(
            utc_time, normalize=False)
        (opos_x, opos_y, opos_z), (ovel_x, ovel_y, ovel_z) = \
            astronomy.observer_position(utc_time, lon, lat, alt)

        lon = np.deg2rad(lon)
        lat = np.deg2rad(lat)

        theta = (astronomy.gmst(utc_time) + lon) % (2 * np.pi)

        rx = pos_x - opos_x
        ry = pos_y - opos_y
        rz = pos_z - opos_z

        sin_lat = np.sin(lat)
        cos_lat = np.cos(lat)
        sin_theta = np.sin(theta)
        cos_theta = np.cos(theta)

        top_s = sin_lat * cos_theta * rx + \
            sin_lat * sin_theta * ry - cos_lat * rz
        top_e = -sin_theta * rx + cos_theta * ry
        top_z = cos_lat * cos_theta * rx + \
            cos_lat * sin_theta * ry + sin_lat * rz

        az_ = np.arctan(-top_e / top_s)

        az_ = np.where(top_s > 0, az_ + np.pi, az_)
        az_ = np.where(az_ < 0, az_ + 2 * np.pi, az_)

        rg_ = np.sqrt(rx * rx + ry * ry + rz * rz)
        el_ = np.arcsin(top_z / rg_)

        return np.rad2deg(az_), np.rad2deg(el_)

    def get_orbit_number(self, utc_time, tbus_style=False, as_float=False):
        """Calculate orbit number at specified time.

        Args:
            utc_time: UTC time as a datetime.datetime object.
            tbus_style: If True, use TBUS-style orbit numbering (TLE orbit number + 1)
            as_float: Return a continuous orbit number as float.
        """
        utc_time = np.datetime64(utc_time)
        try:
            dt = astronomy._days(utc_time - self.orbit_elements.an_time)
            orbit_period = astronomy._days(self.orbit_elements.an_period)
        except AttributeError:
            pos_epoch, vel_epoch = self.get_position(self.tle.epoch,
                                                     normalize=False)
            if np.abs(pos_epoch[2]) > 1 or not vel_epoch[2] > 0:
                # Epoch not at ascending node
                self.orbit_elements.an_time = self.get_last_an_time(
                    self.tle.epoch)
            else:
                # Epoch at ascending node (z < 1 km) and positive v_z
                self.orbit_elements.an_time = self.tle.epoch

            self.orbit_elements.an_period = self.orbit_elements.an_time - \
                self.get_last_an_time(self.orbit_elements.an_time
                                      - np.timedelta64(10, "m"))

            dt = astronomy._days(utc_time - self.orbit_elements.an_time)
            orbit_period = astronomy._days(self.orbit_elements.an_period)

        orbit = self.tle.orbit + dt / orbit_period + \
            self.tle.mean_motion_derivative * dt ** 2 + \
            self.tle.mean_motion_sec_derivative * dt ** 3
        if not as_float:
            orbit = int(orbit)

        if tbus_style:
            orbit += 1

        return orbit

    def get_next_passes(self, utc_time, length, lon, lat, alt, tol=0.001, horizon=0):
        """Calculate passes for the next hours for a given start time and a given observer.

        Original by Martin.

        :utc_time: Observation time (datetime object)
        :length: Number of hours to find passes (int)
        :lon: Longitude of observer position on ground (float)
        :lat: Latitude of observer position on ground (float)
        :alt: Altitude above sea-level (geoid) in km of observer position on ground (float)
        :tol: precision of the result in seconds
        :horizon: the elevation of horizon to compute risetime and falltime.

        :return: [(rise-time, fall-time, max-elevation-time), ...]

        """
        # every minute
        times = utc_time + np.array([dt.timedelta(minutes=minutes)
                                     for minutes in range(length * 60)])
        elev = self.get_observer_look(times, lon, lat, alt)[1] - horizon
        zcs = np.where(np.diff(np.sign(elev)))[0]
        res = []
        risetime = None
        risemins = None
        elev_func = partial(self._elevation, utc_time, lon, lat, alt, horizon)
        elev_inv_func = partial(self._elevation_inv, utc_time, lon, lat, alt, horizon)
        for guess in zcs:
            horizon_mins = _get_root(elev_func, guess, guess + 1.0, tol=tol / 60.0)
            horizon_time = utc_time + dt.timedelta(minutes=horizon_mins)
            if elev[guess] < 0:
                risetime = horizon_time
                risemins = horizon_mins
            else:
                falltime = horizon_time
                fallmins = horizon_mins
                if risetime is None:
                    continue
                int_start = max(0, int(np.floor(risemins)))
                int_end = min(len(elev), int(np.ceil(fallmins) + 1))
                middle = int_start + np.argmax(elev[int_start:int_end])
                highest = utc_time + \
                    dt.timedelta(minutes=_get_max_parab(
                        elev_inv_func,
                        max(risemins, middle - 1), min(fallmins, middle + 1),
                        tol=tol / 60.0
                    ))
                res += [(risetime, falltime, highest)]
        return res

    def _get_time_at_horizon(self, utc_time, obslon, obslat, **kwargs):
        """Determine when the satellite is at the horizon relative to an observer on ground.

        Get the time closest in time to *utc_time* when the satellite is at the
        horizon relative to the position of an observer on ground (altitude =
        0).

        Note: This is considered deprecated and it's functionality is currently
        replaced by 'get_next_passes'.

        """
        warnings.warn("_get_time_at_horizon is replaced with get_next_passes",
                      DeprecationWarning, stacklevel=2)
        if "precision" in kwargs:
            precision = kwargs["precision"]
        else:
            precision = dt.timedelta(seconds=0.001)
        if "max_iterations" in kwargs:
            nmax_iter = kwargs["max_iterations"]
        else:
            nmax_iter = 100

        sec_step = 0.5
        t_step = dt.timedelta(seconds=sec_step / 2.0)

        # Local derivative:
        def fprime(timex):
            el0 = self.get_observer_look(timex - t_step,
                                         obslon, obslat, 0.0)[1]
            el1 = self.get_observer_look(timex + t_step,
                                         obslon, obslat, 0.0)[1]
            return el0, (abs(el1) - abs(el0)) / sec_step

        tx0 = utc_time - dt.timedelta(seconds=1.0)
        tx1 = utc_time
        idx = 0
        # eps = 500.
        eps = 100.
        while abs(tx1 - tx0) > precision and idx < nmax_iter:
            tx0 = tx1
            fpr = fprime(tx0)
            # When the elevation is high the scale is high, and when
            # the elevation is low the scale is low
            # var_scale = np.abs(np.sin(fpr[0] * np.pi/180.))
            # var_scale = np.sqrt(var_scale)
            var_scale = np.abs(fpr[0])
            tx1 = tx0 - dt.timedelta(seconds=(eps * var_scale * fpr[1]))
            idx = idx + 1
            # print idx, tx0, tx1, var_scale, fpr
            if abs(tx1 - utc_time) < precision and idx < 2:
                tx1 = tx1 + dt.timedelta(seconds=1.0)

        if abs(tx1 - tx0) <= precision and idx < nmax_iter:
            return tx1
        else:
            return None

    def utc2local(self, utc_time):
        """Convert UTC to local time."""
        lon, _, _ = self.get_lonlatalt(utc_time)
        return utc_time + dt.timedelta(hours=lon * 24 / 360.0)

    def get_equatorial_crossing_time(self, tstart, tend, node="ascending", local_time=False,
                                     rtol=1E-9):
        """Estimate the equatorial crossing time of an orbit.

        The crossing time is determined via the orbit number, which increases by one if the
        spacecraft passes the ascending node at the equator. A bisection algorithm is used to find
        the time of that passage.

        Args:
            tstart: Start time of the orbit
            tend: End time of the orbit. Orbit number at the end must be at least one greater than
                at the start. If there are multiple revolutions in the given time interval, the
                crossing time of the last revolution in that interval will be computed.
            node: Specifies whether to compute the crossing time at the ascending or descending
                node. Choices: ('ascending', 'descending').
            local_time: By default the UTC crossing time is returned. Use this flag to convert UTC
                to local time.
            rtol: Tolerance of the bisection algorithm. The smaller the tolerance, the more accurate
                the result.
        """
        # Determine orbit number at the start and end of the orbit.
        n_start = self.get_orbit_number(tstart, as_float=True)
        n_end = self.get_orbit_number(tend, as_float=True)
        if int(n_end) - int(n_start) == 0:
            # Orbit doesn't cross the equator in the given time interval
            return None
        elif n_end - n_start > 1:
            warnings.warn("Multiple revolutions between start and end time. Computing crossing "
                          "time for the last revolution in that interval.", stacklevel=2)

        # Let n'(t) = n(t) - offset. Determine offset so that n'(tstart) < 0 and n'(tend) > 0 and
        # n'(tcross) = 0.
        offset = int(n_end)
        if node == "descending":
            offset = offset + 0.5

        # Use bisection algorithm to find the root of n'(t), which is the crossing time. The
        # algorithm requires continuous time coordinates, so convert timestamps to microseconds
        # since 1970.
        time_unit = "us"  # same precision as datetime

        def _nprime(time_f):
            """Continuous orbit number as a function of time."""
            time64 = np.datetime64(int(time_f), time_unit)
            n = self.get_orbit_number(time64, as_float=True)
            return n - offset

        try:
            tcross = optimize.bisect(_nprime,
                                     a=np.datetime64(tstart, time_unit).astype(np.int64),
                                     b=np.datetime64(tend, time_unit).astype(np.int64),
                                     rtol=rtol)
        except ValueError:
            # Bisection did not converge
            return None
        tcross = np.datetime64(int(tcross), time_unit).astype(dt.datetime)

        # Convert UTC to local time
        if local_time:
            tcross = self.utc2local(tcross)

        return tcross


    def _elevation(self, utc_time, lon, lat, alt, horizon, minutes):
        """Compute the elevation."""
        return self.get_observer_look(utc_time +
                                      dt.timedelta(minutes=np.float64(minutes)),
                                      lon, lat, alt)[1] - horizon


    def _elevation_inv(self, utc_time, lon, lat, alt, horizon, minutes):
        """Compute the inverse of elevation."""
        return -self._elevation(utc_time, lon, lat, alt, horizon, minutes)


def _get_root(fun, start, end, tol=0.01):
    """Root finding scheme."""
    x_0 = end
    x_1 = start
    fx_0 = fun(end)
    fx_1 = fun(start)
    if abs(fx_0) < abs(fx_1):
        fx_0, fx_1 = fx_1, fx_0
        x_0, x_1 = x_1, x_0

    x_n = optimize.brentq(fun, x_0, x_1)
    return x_n


def _get_max_parab(fun, start, end, tol=0.01):
    """Successive parabolic interpolation."""
    a = float(start)
    c = float(end)
    b = (a + c) / 2.0

    f_a = fun(a)
    f_b = fun(b)
    f_c = fun(c)

    x = b
    with np.errstate(invalid="raise"):
        while True:
            try:
                x = x - 0.5 * (((b - a) ** 2 * (f_b - f_c)
                                - (b - c) ** 2 * (f_b - f_a)) /
                               ((b - a) * (f_b - f_c) - (b - c) * (f_b - f_a)))
            except FloatingPointError:
                return b
            if abs(b - x) <= tol:
                return x
            f_x = fun(x)
            # sometimes the estimation diverges... return best guess
            if f_x > f_b:
                logger.info("Parabolic interpolation did not converge, returning best guess so far.")
                return b

            a, b, c = (a + x) / 2.0, x, (x + c) / 2.0
            f_a, f_b, f_c = fun(a), f_x, fun(c)


class OrbitElements:
    """Class holding the orbital elements."""

    def __init__(self, tle):
        """Initialize the class."""
        self.epoch = tle.epoch
        self.eccentricity = tle.eccentricity
        self.inclination = np.deg2rad(tle.inclination)
        self.right_ascension = np.deg2rad(tle.right_ascension)
        self.arg_perigee = np.deg2rad(tle.arg_perigee)
        self.mean_anomaly = np.deg2rad(tle.mean_anomaly)

        self.mean_motion = tle.mean_motion * (2 * np.pi / XMNPDA)
        self.mean_motion_derivative = tle.mean_motion_derivative * (2 * np.pi / XMNPDA**2)
        self.mean_motion_sec_derivative = tle.mean_motion_sec_derivative * (2 * np.pi / XMNPDA**3)
        self.bstar = tle.bstar * AE

        self.original_mean_motion, self.semi_major_axis = self._calculate_mean_motion_and_semi_major_axis()

        self.period = 2 * np.pi / self.original_mean_motion
        self.right_ascension_lon = self.right_ascension - astronomy.gmst(self.epoch)
        self.right_ascension_lon = np.fmod(self.right_ascension_lon + np.pi, 2 * np.pi) - np.pi

    @property
    def excentricity(self):
        """Get 'eccentricity' using legacy 'excentricity' name."""
        warnings.warn("The 'eccentricity' property is deprecated in favor of 'eccentricity'", stacklevel=2)
        return self.eccentricity

    @property
    def apogee(self):
        """Compute apogee altitude in kilometers."""
        return ((self.semi_major_axis * (1 + self.eccentricity)) / AE - AE) * XKMPER

    @property
    def perigee(self):
        """Compute perigee altitude in kilometers."""
        return ((self.semi_major_axis * (1 - self.eccentricity)) / AE - AE) * XKMPER

    @property
    def is_circular(self):
        """Check if orbit is nearly circular."""
        return self.eccentricity < 1e-3

    @property
    def is_retrograde(self):
        """Check if orbit is retrograde (inclination > 90°)."""
        return self.inclination > np.pi / 2

    def _get_true_anomaly(self):
        """Computes the True Anomaly (nu) from Mean Anomaly (M) and Eccentricity (e)."""
        M = self.mean_anomaly
        e = self.eccentricity
        E = M # Initial guess for Eccentric Anomaly (E)

        # Iteratively solve Kepler's Equation (M = E - e*sin(E))
        for _ in range(10): # Max 10 iterations
            f = E - e * np.sin(E) - M
            f_prime = 1 - e * np.cos(E)
            E_new = E - f / f_prime
            if np.abs(E_new - E) < 1e-8: # Tolerance
                E = E_new
                break
            E = E_new

        # Convert Eccentric Anomaly (E) to True Anomaly (nu)
        nu = 2 * np.arctan2(
            np.sqrt(1 + e) * np.sin(E / 2),
            np.sqrt(1 - e) * np.cos(E / 2)
        )
        return nu

    def position_vector_in_orbital_plane(self):
        """Compute position vector in the orbital plane at epoch.

        The x-axis points toward the perigee.
        """
        true_anomaly = self._get_true_anomaly()

        # Calculate radius (r) and coordinates using the True Anomaly
        r = self.semi_major_axis * (1 - self.eccentricity**2) / \
            (1 + self.eccentricity * np.cos(true_anomaly))

        x = r * np.cos(true_anomaly)
        y = r * np.sin(true_anomaly)

        return np.array([x, y])

    def _get_velocity_at_apsis(self, e_1, e_2):
        """Helper method to compute orbital velocity at an apsis."""
        mu = XKE**2 * AE**3

        # r_apsis is the distance at the apsis (perigee or apogee)
        r_apsis = self.semi_major_axis * e_1

        # The velocity formula: V = sqrt( mu * (2/r - 1/a) ).
        # For apse lines (r = a * (1-e^2)/(1+e*cos(nu)), it simplifies to:
        # V = sqrt( mu * (1 ± e) / r )
        v_er_per_min = np.sqrt(mu * (e_2 / r_apsis))

        # Conversion factor: (AE * XKMPER) converts ER -> km; / 60 converts min -> s
        conversion_factor = (AE * XKMPER) / 60.0

        return v_er_per_min * conversion_factor

    def velocity_at_perigee(self):
        """Compute orbital velocity at perigee in km/s."""
        return self._get_velocity_at_apsis(1 - self.eccentricity, 1 + self.eccentricity)

    def velocity_at_apogee(self):
        """Compute orbital velocity at apogee in km/s."""
        return self._get_velocity_at_apsis(1 + self.eccentricity, 1 - self.eccentricity)

    def _calculate_mean_motion_and_semi_major_axis(self):
        """Apply SGP4 perturbation corrections to mean motion and semi-major axis.

        Based on inclination and eccentricity.
        """
        a_1 = (XKE / self.mean_motion) ** (2.0 / 3)
        delta_1 = (3 / 2.0) * (CK2 / a_1**2) * ((3 * np.cos(self.inclination)**2 - 1) /
                                               (1 - self.eccentricity**2)**(3 / 2))
        a_0 = a_1 * (1 - delta_1 / 3 - delta_1**2 - (134.0 / 81) * delta_1**3)
        delta_0 = (3 / 2.0) * (CK2 / a_0**2) * ((3 * np.cos(self.inclination)**2 - 1) /
                                               (1 - self.eccentricity**2)**(3 / 2))

        corrected_mean_motion = self.mean_motion / (1 + delta_0)
        corrected_semi_major_axis = a_0 / (1 - delta_0)

        return corrected_mean_motion, corrected_semi_major_axis


class _SGDP4Base:
    """Helper class for the SGDP4 computations."""

    def __init__(self, orbit_elements):
        """Initialize class."""
        self.mode = None

        _check_orbital_elements(orbit_elements)

        self.eo = orbit_elements.eccentricity
        self.xincl = orbit_elements.inclination
        self.xno = orbit_elements.original_mean_motion
        self.bstar = orbit_elements.bstar
        self.omegao = orbit_elements.arg_perigee
        self.xmo = orbit_elements.mean_anomaly
        self.xnodeo = orbit_elements.right_ascension
        self.t_0 = orbit_elements.epoch
        self.xn_0 = orbit_elements.mean_motion

        if self.eo < 0:
            self.mode = self.SGDP4_ZERO_ECC
            return

        self.cosIO = np.cos(self.xincl)
        self.sinIO = np.sin(self.xincl)
        theta2 = self.cosIO**2
        self.x3thm1 = 3.0 * theta2 - 1.0
        self.x1mth2 = 1.0 - theta2
        self.x7thm1 = 7.0 * theta2 - 1.0

        self.xnodp = None
        self.aodp = None
        self.perigee = None
        self.apogee = None
        self.period = None
        self._betao = None
        self._betao2 = None
        self._calculate_basic_orbit_params()

        self._set_mode()

        s4, qoms24 = self._get_s4_qoms24()
        tsi = 1.0 / (self.aodp - s4)
        self.eta = self.aodp * self.eo * tsi
        eeta = self.eo * self.eta
        coef = qoms24 * tsi**4

        self.c1 = None
        self.c2 = None
        self.c3 = None
        self.c4 = None
        self.c5 = None
        self._calculate_c_coefficients(coef, eeta, tsi)

        self.xmdot = None
        self.omgdot = None
        self.xnodot = None
        self._calculate_dot_products(theta2)

        self.xmcof = self._calculate_xmcof(coef, eeta)
        self.xnodcf = 3.5 * self._betao2 * self._xhdot1 * self.c1
        self.t2cof = 1.5 * self.c1
        self.xlcof = self._calculate_xlcof()
        self.aycof = 0.25 * A3OVK2 * self.sinIO
        self.cosXMO = np.cos(self.xmo)
        self.sinXMO = np.sin(self.xmo)
        self.delmo = (1.0 + self.eta * self.cosXMO)**3

        self.d2 = None
        self.d3 = None
        self.d4 = None
        self.t3cof = None
        self.t4cof = None
        self.t5cof = None
        if self.mode == SGDP4_NEAR_NORM:
            self._calculate_near_norm_parameters(tsi, s4)
        elif self.mode == SGDP4_DEEP_NORM:
            raise NotImplementedError("Deep space calculations not supported")

    def _calculate_basic_orbit_params(self):
        a1 = (XKE / self.xn_0) ** (2. / 3)
        self._betao2 = 1.0 - self.eo**2
        self._betao = np.sqrt(self._betao2)
        temp0 = 1.5 * CK2 * self.x3thm1 / (self._betao * self._betao2)
        del1 = temp0 / (a1**2)
        a0 = a1 * (1.0 - del1 * (1.0 / 3.0 + del1 * (1.0 + del1 * 134.0 / 81.0)))
        del0 = temp0 / (a0**2)
        self.xnodp = self.xn_0 / (1.0 + del0)
        self.aodp = (a0 / (1.0 - del0))
        self.perigee = (self.aodp * (1.0 - self.eo) - AE) * XKMPER
        self.apogee = (self.aodp * (1.0 + self.eo) - AE) * XKMPER
        self.period = (2 * np.pi * 1440.0 / XMNPDA) / self.xnodp

    def _set_mode(self):
        if self.period >= 225:
            # Deep-Space model
            self.mode = SGDP4_DEEP_NORM
        elif self.perigee < 220:
            # Near-space, simplified equations
            self.mode = SGDP4_NEAR_SIMP
        else:
            # Near-space, normal equations
            self.mode = SGDP4_NEAR_NORM

    def _calculate_c_coefficients(self, coef, eeta, tsi):
        etasq = self.eta**2
        psisq = np.abs(1.0 - etasq)
        coef_1 = coef / psisq**3.5
        self.c2 = (coef_1 * self.xnodp * (self.aodp *
                                          (1.0 + 1.5 * etasq + eeta * (4.0 + etasq)) +
                                          (0.75 * CK2) * tsi / psisq * self.x3thm1 *
                                          (8.0 + 3.0 * etasq * (8.0 + etasq))))

        self.c1 = self.bstar * self.c2

        self.c4 = (2.0 * self.xnodp * coef_1 * self.aodp * self._betao2 * (
            self.eta * (2.0 + 0.5 * etasq) + self.eo * (0.5 + 2.0 * etasq) - (2.0 * CK2) * tsi /
            (self.aodp * psisq) * (-3.0 * self.x3thm1 * (1.0 - 2.0 * eeta + etasq * (1.5 - 0.5 * eeta)) +
                                   0.75 * self.x1mth2 * (2.0 * etasq - eeta * (1.0 + etasq)) *
                                   np.cos(2.0 * self.omegao))))

        self.c5, self.c3, self.omgcof = 0.0, 0.0, 0.0

        if self.mode == SGDP4_NEAR_NORM:
            self.c5 = (2.0 * coef_1 * self.aodp * self._betao2 *
                       (1.0 + 2.75 * (etasq + eeta) + eeta * etasq))
            if self.eo > ECC_ALL:
                self.c3 = coef * tsi * A3OVK2 * \
                    self.xnodp * AE * self.sinIO / self.eo
            self.omgcof = self.bstar * self.c3 * np.cos(self.omegao)

    def _get_s4_qoms24(self):
        if self.perigee < 156:
            s4 = self.perigee - 78
            if s4 < 20:
                s4 = 20

            qoms24 = ((120 - s4) * (AE / XKMPER))**4
            s4 = (s4 / XKMPER + AE)
            return (s4, qoms24)
        return (KS, QOMS2T)

    def _calculate_dot_products(self, theta2):
        pinvsq = 1.0 / (self.aodp**2 * self._betao2**2)
        temp1 = 3.0 * CK2 * pinvsq * self.xnodp
        temp2 = temp1 * CK2 * pinvsq
        temp3 = 1.25 * CK4 * pinvsq**2 * self.xnodp
        theta4 = theta2 ** 2

        self.xmdot = (self.xnodp + (0.5 * temp1 * self._betao * self.x3thm1 + 0.0625 *
                                    temp2 * self._betao * (13.0 - 78.0 * theta2 +
                                                     137.0 * theta4)))

        x1m5th = 1.0 - 5.0 * theta2

        self.omgdot = (-0.5 * temp1 * x1m5th + 0.0625 * temp2 *
                       (7.0 - 114.0 * theta2 + 395.0 * theta4) +
                       temp3 * (3.0 - 36.0 * theta2 + 49.0 * theta4))

        self._xhdot1 = -temp1 * self.cosIO
        self.xnodot = (self._xhdot1 + (0.5 * temp2 * (4.0 - 19.0 * theta2) +
                                 2.0 * temp3 * (3.0 - 7.0 * theta2)) * self.cosIO)

    def _calculate_xmcof(self, coef, eeta):
        if self.eo > ECC_ALL:
            return (-(2. / 3) * AE) * coef * self.bstar / eeta
        return 0.0

    def _calculate_xlcof(self):
        # Check for possible divide-by-zero for X/(1+cos(xincl)) when
        # calculating xlcof */
        temp0 = 1.0 + self.cosIO
        if np.abs(temp0) < EPS_COS:
            temp0 = np.sign(temp0) * EPS_COS
        return 0.125 * A3OVK2 * self.sinIO * (3.0 + 5.0 * self.cosIO) / temp0

    def _calculate_near_norm_parameters(self, tsi, s4):
        c1sq = self.c1**2
        self.d2 = 4.0 * self.aodp * tsi * c1sq
        temp0 = self.d2 * tsi * self.c1 / 3.0
        self.d3 = (17.0 * self.aodp + s4) * temp0
        self.d4 = 0.5 * temp0 * self.aodp * tsi * \
            (221.0 * self.aodp + 31.0 * s4) * self.c1
        self.t3cof = self.d2 + 2.0 * c1sq
        self.t4cof = 0.25 * \
            (3.0 * self.d3 + self.c1 * (12.0 * self.d2 + 10.0 * c1sq))
        self.t5cof = (0.2 * (3.0 * self.d4 + 12.0 * self.c1 * self.d3 + 6.0 * self.d2**2 +
                                15.0 * c1sq * (2.0 * self.d2 + c1sq)))


class _SGDP4:
    """Class for SGDP4 computations."""

    def __init__(self, orbital_elements):
        self._params = _SGDP4Base(orbital_elements)

    @property
    def eo(self):
        return self._params.eo

    @property
    def xincl(self):
        return self._params.xincl

    @property
    def xno(self):
        return self._params.xno

    @property
    def bstar(self):
        return self._params.bstar

    @property
    def omegao(self):
        return self._params.omegao

    @property
    def xmo(self):
        return self._params.xmo

    @property
    def xnodeo(self):
        return self._params.xnodeo

    @property
    def t_0(self):
        return self._params.t_0

    @property
    def xn_0(self):
        return self._params.xn_0

    @property
    def mode(self):
        return self._params.mode

    @property
    def cosIO(self):
        return self._params.cosIO

    @property
    def sinIO(self):
        return self._params.sinIO

    @property
    def x3thm1(self):
        return self._params.x3thm1

    @property
    def x1mth2(self):
        return self._params.x1mth2

    @property
    def x7thm1(self):
        return self._params.x7thm1

    @property
    def xnodp(self):
        return self._params.xnodp

    @property
    def aodp(self):
        return self._params.aodp

    @property
    def perigee(self):
        return self._params.perigee

    @property
    def apogee(self):
        return self._params.apogee

    @property
    def period(self):
        return self._params.period

    @property
    def eta(self):
        return self._params.eta

    @property
    def c1(self):
        return self._params.c1

    @property
    def c2(self):
        return self._params.c2

    @property
    def c3(self):
        return self._params.c3

    @property
    def c4(self):
        return self._params.c4

    @property
    def c5(self):
        return self._params.c5

    @property
    def xmdot(self):
        return self._params.xmdot

    @property
    def omgdot(self):
        return self._params.omgdot

    @property
    def xnodot(self):
        return self._params.xnodot

    @property
    def xmcof(self):
        return self._params.xmcof

    @property
    def xnodcf(self):
        return self._params.xnodcf

    @property
    def t2cof(self):
        return self._params.t2cof

    @property
    def xlcof(self):
        return self._params.xlcof

    @property
    def aycof(self):
        return self._params.aycof

    @property
    def cosXMO(self):
        return self._params.cosXMO

    @property
    def sinXMO(self):
        return self._params.sinXMO

    @property
    def delmo(self):
        return self._params.delmo

    @property
    def d2(self):
        return self._params.d2

    @property
    def d3(self):
        return self._params.d3

    @property
    def d4(self):
        return self._params.d4

    @property
    def t3cof(self):
        return self._params.t3cof

    @property
    def t4cof(self):
        return self._params.t4cof

    @property
    def t5cof(self):
        return self._params.t5cof


    def propagate(self, utc_time):
        if self.mode == SGDP4_ZERO_ECC:
            raise NotImplementedError("Mode SGDP4_ZERO_ECC not implemented")
        elif self.mode != SGDP4_NEAR_NORM:
            raise NotImplementedError("Deep space calculations not supported")

        kep = _Keplerians(self._params)
        return kep.calculate(utc_time)


class _Keplerians:
    """Class for computing Keplerian parameters."""

    def __init__(self, params):
        """Initialize the class."""
        self._params = params
        self.ecc = None
        self.radius = None
        self.theta = None
        self.eqinc = None
        self.ascn = None
        self.argp = None
        self.smjaxs = None
        self.rdotk = None
        self.rfdotk = None
        self.omega = None

        self._utc_time = None
        self._ts = None
        self._xmp = None
        self._xnode = None
        self._temp0 = None
        self._elsq = None
        self._tempe = None
        self._templ = None
        self._a = None
        self._axn = None
        self._ayn = None
        self._xlt = None
        self._betal = None
        self._pl = None
        self._r = None
        self._ecosE = None
        self._invR = None
        self._cosEPW = None
        self._esinE = None
        self._sinEPW = None
        self._u = None
        self._sin2u = None
        self._cos2u = None
        self._temp1 = None
        self._temp2 = None

    def calculate(self, utc_time):
        """Calculate Keplerians and return them as a dict."""
        self._utc_time = utc_time
        self._get_timedelta_in_minutes()

        self._xmp = self._params.xmo + self._params.xmdot * self._ts
        self._xnode = self._params.xnodeo + self._ts * (self._params.xnodot + self._ts * self._params.xnodcf)

        delm = self._params.xmcof * \
            ((1.0 + self._params.eta * np.cos(self._xmp))**3 - self._params.delmo)
        self._temp0 = self._ts * self._params.omgcof + delm
        self._xmp += self._temp0

        self._calculate_omega()
        self._calculate_tempe()
        self._calculate_templ()

        self._calculate_a()
        self._calculate_axn_and_ayn()
        self._elsq = _calculate_elsq(self._axn, self._ayn, self._utc_time)

        self.ecc = np.sqrt(self._elsq)

        self._calculate_preliminary_short_period()
        self._update_short_period()
        kep = self._collect_return_values()

        return kep

    def _get_timedelta_in_minutes(self):
        self._ts = (dt2np(self._utc_time) - self._params.t_0) / np.timedelta64(1, "m")

    def _calculate_omega(self):
        self.omega = self._params.omegao + self._params.omgdot * self._ts - self._temp0

    def _calculate_tempe(self):
        if self._params.mode == SGDP4_NEAR_SIMP:
            self._tempe = self._params.bstar * self._ts * self._params.c4
        else:
            self._tempe = self._params.bstar * \
                (self._params.c4 * self._ts + self._params.c5 * (np.sin(self._xmp) - self._params.sinXMO))

    def _calculate_templ(self):
        if self._params.mode == SGDP4_NEAR_SIMP:
            self._templ = self._ts * self._ts * self._params.t2cof
        else:
            self._templ = self._ts * self._ts * \
            (self._params.t2cof + self._ts *
                (self._params.t3cof + self._ts * (self._params.t4cof + self._ts * self._params.t5cof)))

    def _calculate_a(self):
        if self._params.mode == SGDP4_NEAR_SIMP:
            tempa = 1.0 - self._ts * self._params.c1
        else:
            tempa = 1.0 - \
                (self._ts *
                 (self._params.c1 + self._ts * (self._params.d2 + self._ts *
                  (self._params.d3 + self._ts * self._params.d4))))
        self._a = self._params.aodp * tempa**2

        if np.any(self._a < 1):
            raise Exception("Satellite crashed at time %s", self._utc_time)

    def _calculate_axn_and_ayn(self):
        e = self._calculate_e(self._tempe)
        beta2 = 1.0 - e**2

        # Long period periodics
        sinOMG = np.sin(self.omega)
        cosOMG = np.cos(self.omega)

        self._temp0 = 1.0 / (self._a * beta2)
        self._axn = e * cosOMG
        self._ayn = e * sinOMG + self._temp0 * self._params.aycof

    def _calculate_e(self, tempe):
        e = self._params.eo - tempe

        if np.any(e < ECC_LIMIT_LOW):
            raise ValueError("Satellite modified eccentricity too low: %s < %e"
                             % (str(e[e < ECC_LIMIT_LOW]), ECC_LIMIT_LOW))

        e = np.where(e < ECC_EPS, ECC_EPS, e)
        e = np.where(e > ECC_LIMIT_HIGH, ECC_LIMIT_HIGH, e)

        return e

    def _calculate_preliminary_short_period(self):
        xl = self._xmp + self.omega + self._xnode + self._params.xnodp * self._templ
        self._xlt = xl + self._temp0 * self._params.xlcof * self._axn

        self._iterate_newton_raphson()

        # Short period preliminary quantities
        self._temp0 = 1.0 - self._elsq
        self._betal = np.sqrt(self._temp0)
        self._pl = self._a * self._temp0
        self._r = self._a * (1.0 - self._ecosE)
        self._invR = 1.0 / self._r
        temp2 = self._a * self._invR
        temp3 = 1.0 / (1.0 + self._betal)
        cosu = temp2 * (self._cosEPW - self._axn + self._ayn * self._esinE * temp3)
        sinu = temp2 * (self._sinEPW - self._ayn - self._axn * self._esinE * temp3)

        self._u = np.arctan2(sinu, cosu)
        self._sin2u = 2.0 * sinu * cosu
        self._cos2u = 2.0 * cosu**2 - 1.0
        self._temp0 = 1.0 / self._pl
        self._temp1 = CK2 * self._temp0
        self._temp2 = self._temp1 * self._temp0

    def _iterate_newton_raphson(self):
        epw = np.fmod(self._xlt - self._xnode, 2 * np.pi)
        # needs a copy in case of an array
        capu = np.array(epw)
        for i in range(10):
            self._sinEPW = np.sin(epw)
            self._cosEPW = np.cos(epw)

            self._ecosE = self._axn * self._cosEPW + self._ayn * self._sinEPW
            self._esinE = self._axn * self._sinEPW - self._ayn * self._cosEPW
            f = capu - epw + self._esinE
            if np.all(np.abs(f) < NR_EPS):
                break

            df = 1.0 - self._ecosE

            # 1st order Newton-Raphson correction.
            nr = f / df

            # 2nd order Newton-Raphson correction.
            nr = np.where(np.logical_and(i == 0, np.abs(nr) > 1.25 * self.ecc),
                          np.sign(nr) * self.ecc,
                          f / (df + 0.5 * self._esinE * nr))
            epw += nr


    def _update_short_period(self):
        self.rk = self._r * (1.0 - 1.5 * self._temp2 * self._betal * self._params.x3thm1) + \
            0.5 * self._temp1 * self._params.x1mth2 * self._cos2u
        self.uk = self._u - 0.25 * self._temp2 * self._params.x7thm1 * self._sin2u
        self.xnodek = self._xnode + 1.5 * self._temp2 * self._params.cosIO * self._sin2u
        self.xinc = self._params.xincl + 1.5 * self._temp2 * self._params.cosIO * self._params.sinIO * self._cos2u

        if np.any(self.rk < 1):
            raise Exception("Satellite crashed at time %s", self._utc_time)

        self._temp0 = np.sqrt(self._a)
        temp2 = XKE / (self._a * self._temp0)
        self.rdotk = (
            (XKE * self._temp0 * self._esinE * self._invR -
                temp2 * self._temp1 * self._params.x1mth2 * self._sin2u) *
            (XKMPER / AE * XMNPDA / 86400.0))
        self.rfdotk = ((XKE * np.sqrt(self._pl) * self._invR + temp2 * self._temp1 *
                   (self._params.x1mth2 * self._cos2u + 1.5 * self._params.x3thm1)) *
                  (XKMPER / AE * XMNPDA / 86400.0))

    def _collect_return_values(self):
        kep = {}
        kep["ecc"] = self.ecc
        kep["radius"] = self.rk * XKMPER / AE
        kep["theta"] = self.uk
        kep["eqinc"] = self.xinc
        kep["ascn"] = self.xnodek
        kep["argp"] = self.omega
        kep["smjaxs"] = self._a * XKMPER / AE
        kep["rdotk"] = self.rdotk
        kep["rfdotk"] = self.rfdotk

        return kep


def _check_orbital_elements(orbit_elements):
    if not (0 < orbit_elements.eccentricity < ECC_LIMIT_HIGH):
        raise OrbitalError("Eccentricity out of range: %e" % orbit_elements.eccentricity)
    if not ((0.0035 * 2 * np.pi / XMNPDA) < orbit_elements.original_mean_motion < (18 * 2 * np.pi / XMNPDA)):
        raise OrbitalError("Mean motion out of range: %e" % orbit_elements.original_mean_motion)
    if not (0 < orbit_elements.inclination < np.pi):
        raise OrbitalError("Inclination out of range: %e" % orbit_elements.inclination)


def _calculate_elsq(axn, ayn, utc_time):
    elsq = axn**2 + ayn**2

    if np.any(elsq >= 1):
        raise Exception("e**2 >= 1 at %s", utc_time)

    return elsq


def _get_tz_unaware_utctime(utc_time):
    """Return timzone unaware datetime object.

    The input *utc_time* is either a timezone unaware object assumed to be in
    UTC, or a timezone aware datetime object in UTC.
    """
    if isinstance(utc_time, dt.datetime):
        if utc_time.tzinfo and utc_time.tzinfo != dt.timezone.utc:
            raise ValueError("UTC time expected! Parsing a timezone aware datetime object requires it to be UTC!")
        return utc_time.replace(tzinfo=None)

    return utc_time


def kep2xyz(kep):
    """Keppler to cartesian coordinates conversion.

    (Not sure what 'kep' actually refers to, just guessing! FIXME!)
    """
    sinT = np.sin(kep["theta"])
    cosT = np.cos(kep["theta"])
    sinI = np.sin(kep["eqinc"])
    cosI = np.cos(kep["eqinc"])
    sinS = np.sin(kep["ascn"])
    cosS = np.cos(kep["ascn"])

    xmx = -sinS * cosI
    xmy = cosS * cosI

    ux = xmx * sinT + cosS * cosT
    uy = xmy * sinT + sinS * cosT
    uz = sinI * sinT

    x = kep["radius"] * ux
    y = kep["radius"] * uy
    z = kep["radius"] * uz

    vx = xmx * cosT - cosS * sinT
    vy = xmy * cosT - sinS * sinT
    vz = sinI * cosT

    v_x = kep["rdotk"] * ux + kep["rfdotk"] * vx
    v_y = kep["rdotk"] * uy + kep["rfdotk"] * vy
    v_z = kep["rdotk"] * uz + kep["rfdotk"] * vz

    return np.array((x, y, z)), np.array((v_x, v_y, v_z))


if __name__ == "__main__":
    obs_lon, obs_lat = np.deg2rad((12.4143, 55.9065))
    obs_alt = 0.02
    o = Orbital(satellite="METOP-B")

    t_start = dt.datetime.now()
    t_stop = t_start + dt.timedelta(minutes=20)
    t = t_start
    while t < t_stop:
        t += dt.timedelta(seconds=15)
        lon, lat, alt = o.get_lonlatalt(t)
        lon, lat = np.rad2deg((lon, lat))
        az, el = o.get_observer_look(t, obs_lon, obs_lat, obs_alt)
        ob = o.get_orbit_number(t, tbus_style=True)
        print(az, el, ob)
